<?php

/**
 * REST API Routes
 *
 *
 * Make sure to add the lines below to .htaccess
 * otherwise Apache may strip out the auth header.
 * RewriteCond %{HTTP:Authorization} ^(.*)
 * RewriteRule ^(.*) - [E=HTTP_AUTHORIZATION:%1]
 *
 *
 * Add line below to wp-config.php
 * define('MLM_API_SECRET', '78PLqsD]8#hQBdC$)Q$i`1WuD6Fgf2=RjyE:p%&S4}EMa|uW!^m?/Yn|u;3$)1%u');
 *
 *
 * @pack WebHow
 */
class MlmRestRoutes
{
	/**
	 * Global variables.
	 */
	private static $TOKEN_BYTE_LENGTH	= 32;
	private static $MLM_ALG				= 'HS256';
	private static $MLM_SIGNATURE		= 'sha256';
	private static $DATE_FORMAT			= 'Y-m-d H:i:s';

	/**
	 * To keep the initialization fast, only add filter and action
	 * hooks in the constructor.
	 */
	public function __construct()
	{
		add_action( 'rest_api_init', array( $this, 'rest_api_init_action' ) );
		add_action( 'deleted_user', array( $this, 'deleted_user_action' ) );

		add_filter( 'manage_users_columns', array( $this, 'manage_users_columns_filter' ) );
		add_filter( 'manage_users_custom_column', array( $this, 'manage_users_custom_column_filter' ), 10, 3 );
		add_filter( 'bulk_actions-users', array( $this, 'bulk_actions_edit_users_filter' ) );
		add_filter( 'handle_bulk_actions-users', array( $this, 'handle_bulk_actions_users' ), 10, 3 );
	}


	/*
	 * Returns WP user id from access token.
	 * @param $token Access token of user
	 * @return integer|null WP user is or null if nothing is found.
	 */
	public function get_user_id_from_access_token( $token )
	{
		global $wpdb;
		$table	= $wpdb->prefix . 'mlm_token';

		if( ! defined('MLM_API_SECRET') )
		{
			$result	= $wpdb->get_var( $wpdb->prepare(
				"SELECT user_id FROM {$table} WHERE access_token = %s AND access_token_valid > NOW()",
				$token
			) );

			return $result;
		}

		$payload	= $this->verify_token( $token );

		if( is_wp_error( $payload ) )
		{
			return $payload;
		}

		return $payload->sub;
	}


	/*
	 * Returns token info from user id.
	 * @param $user_id WP user id
	 * @param $column_name token table column name
	 * @return string|null WP user is or null if nothing is found.
	 */
	public function get_user_token_info( $user_id, $column_name )
	{
		global $wpdb;
		$table	= $wpdb->prefix . 'mlm_token';
		$result	= $wpdb->get_var( $wpdb->prepare(
			"SELECT {$column_name} FROM {$table} WHERE user_id = %d LIMIT 1",
			$user_id
		) );

		return $result;
	}


	/**
	 * Returns user id for this refresh token.
	 * @param string $token Refresh token.
	 * @return integer|null User id or null when nothing is found.
	 */
	public function get_user_id_from_refresh_token( $token )
	{
		global $wpdb;
		$table	= $wpdb->prefix . 'mlm_token';
		$result	= $wpdb->get_var( $wpdb->prepare(
			"SELECT user_id FROM {$table} WHERE refresh_token = %s",
			$token
		) );

		return $result;
	}


	/**
	 * Remove token for this user.
	 * @param integer $id User id.
	 * @return void
	 */
	public function delete_by_user_id( $id )
	{
		global $wpdb;
		$table	= $wpdb->prefix . 'mlm_token';
		$wpdb->delete( $table, array( 'user_id' => $id ) );
	}


	/**
	 * Refresh the access token for the user that belongs to this refresh token.
	 * @param $refresh_token Refresh token of this user.
	 * @return array|false Array with new access token or false on error.
	 */
	public function refresh_access_token( $refresh_token, $user_id )
	{
		global $wpdb;
		$table	= $wpdb->prefix . 'mlm_token';
		$token	= $this->create_access_token( $user_id );
		$result	= $wpdb->query( $wpdb->prepare(
			"UPDATE {$table} SET access_token = %s, access_token_valid = %s",
			$token['token'],
			$token['expires_datetime']
		) );

		if( $result )
		{
			return array(
				'access_token'	=> $token['token'],
				'expires_in'	=> $token['expires_in']
			);
		}

		return false;
	}


	/*
	 * Inserts or update access and refresh tokens for user after login.
	 * @param $user_id ID of WP user
	 * @return false on error or array with access and refresh token on success
	 */
	public function after_login( $user_id )
	{
		global $wpdb;
		$table			= $wpdb->prefix . 'mlm_token';
		$access_token	= $this->create_access_token( $user_id );
		$refresh_token	= $this->create_refresh_token( $user_id );
		$result			= $wpdb->query( $wpdb->prepare(
			"INSERT INTO {$table} SET user_id = %d, access_token = %s, access_token_valid = %s, refresh_token = %s
			ON DUPLICATE KEY UPDATE access_token = %s, access_token_valid = %s, refresh_token = %s",
			$user_id,
			$access_token['token'],
			$access_token['expires_datetime'],
			$refresh_token,
			$access_token['token'],
			$access_token['expires_datetime'],
			$refresh_token
		) );

		if( $result !== false )
		{
			return array(
				'access_token'	=> $access_token['token'],
				'expires_in'	=> $access_token['expires_in'],
				'refresh_token'	=> $refresh_token
			);
		}

		// Error
		return false;
	}


	/*
	 * Returns access token.
	 */
	private function create_access_token( $user_id )
	{
		$expire		= 2592000; // one month
		$exp		= new DateTime();
		$nowTs		= $exp->getTimeStamp();
		$exp->add( new DateInterval( 'PT' . $expire . 'S' ) );
		$expTs		= $exp->getTimeStamp();
		$expiresIn	= $expTs - $nowTs;
		$token		= null;

		if( ! defined('MLM_API_SECRET') )
		{
			$token	= bin2hex(openssl_random_pseudo_bytes( self::$TOKEN_BYTE_LENGTH ) );
		}
		else
		{
			$header				= $this->base64url_encode( json_encode( array( 'typ' => 'JWT', 'alg' => self::$MLM_ALG ) ) );
			$payload			= $this->base64url_encode( json_encode( array( 'exp' => $expTs, 'sub' => $user_id ) ) );
			$dataEncoded		= $header . '.' . $payload;
			$signature			= hash_hmac( self::$MLM_SIGNATURE, $dataEncoded, MLM_API_SECRET, true );
			$signatureEncoded	= $this->base64url_encode( $signature );
			$token				= $dataEncoded . '.' . $signatureEncoded;
		}

		return array(
			'token'				=> $token,
			'expires_datetime'	=> $exp->format( self::$DATE_FORMAT ),
			'expires_in'		=> $expiresIn
		);
	}


	/**
	 * Verify the given token.
	 * @param string $token Access token.
	 */
	private function verify_token( $token )
	{
		$parts	= explode( '.', $token );

		if( count( $parts ) !== 3 )
		{
			return $this->get_error('wrong_token');
		}

		// $header = json_decode( $this->base64url_decode( $parts[0] ) );
		$result = $this->base64url_encode( hash_hmac( self::$MLM_SIGNATURE, $parts[0] . '.' . $parts[1], MLM_API_SECRET, true ) );

		if( $result !== $parts[2] )
		{
			return $this->get_error('signature_error');
		}

		$payload	= json_decode( $this->base64url_decode( $parts[1] ) );

		if( $payload->exp < time() )
		{
			return $this->get_error('token_expired');
		}

		return $payload;
	}


	/*
	 * Base64 encode string.
	 */
	private function base64url_encode( $data )
	{
		return rtrim(strtr(base64_encode($data), '+/', '-_'), '=');
	}


	/*
	 * Base64 decode string.
	 */
	private function base64url_decode( $data )
	{
		return base64_decode(strtr($data, '-_', '+/'));
	}


	/*
	 * Returns refresh token.
	 */
	private function create_refresh_token( $user_id )
	{
		return bin2hex(openssl_random_pseudo_bytes( self::$TOKEN_BYTE_LENGTH ));
	}


	/*
	 * Insert a new bulk action.
	 */
	public function bulk_actions_edit_users_filter( $bulk_actions )
	{
		$bulk_actions['mlm_revoke_tokens']	= 'حذف توکن وب سرویس';

		return $bulk_actions;
	}


	/*
	 * Handle revoke token bulk action.
	 */
	public function handle_bulk_actions_users( $redirect_to, $doaction, $user_ids )
	{
		if( $doaction !== 'mlm_revoke_tokens' )
		{
			return $redirect_to;
		}

		foreach( $user_ids as $user_id )
		{
			$this->delete_by_user_id( $user_id );
		}

		$redirect_to	= add_query_arg( 'mlm_bulk_revoked_tokens', count( $user_ids ), $redirect_to );

		return $redirect_to;
	}


	/*
	 * Insert new column for users table.
	 */
	public function manage_users_columns_filter( $column_headers )
	{
		$column_headers['mlm_token_ex']	= 'اعتبار توکن تا';

		return $column_headers;
	}


	/*
	 * Token expire column value.
	 */
	public function manage_users_custom_column_filter( $output, $column_name, $user_id )
	{
		switch( $column_name )
		{
			case 'mlm_token_ex':
				$s	= $this->get_user_token_info( $user_id, 'access_token_valid' );

				if( ! empty( $s ) )
				{
					return date_i18n( 'j F Y', strtotime( $s ) );
				}

				return '-';

			default:
				return $output;
		}
	}


	/**
	 * Action fired when user is deleted. Deletes tokens for this user.
	 * @param integer $id User id.
	 */
	public function deleted_user_action( $id )
	{
		$this->delete_by_user_id( $id );
	}


	/**
	 * Find current user id.
	 */
	private function find_user_id_by_token()
	{
		$headers		= function_exists('apache_request_headers') ? apache_request_headers() : $_SERVER;
		$possibleKeys	= ['Authorization', 'HTTP_AUTHORIZATION', 'REDIRECT_HTTP_AUTHORIZATION'];
		$authHeader		= null;

		foreach( $possibleKeys as $key )
		{
			if( ! empty( $headers[$key] ) )
			{
				$authHeader	= $headers[$key];
				break;
			}
		}

		if( ! empty( $authHeader ) )
		{
			// 7 = strlen('Bearer ');
			$access_token	= substr( $authHeader, 7 );
			$my_user_id		= $this->get_user_id_from_access_token( $access_token );

			if( ! is_wp_error( $my_user_id ) )
			{
				return $my_user_id;
			}
		}

		return false;
	}


	/**
	 * Registering the routes endpoints
	 */
	public function rest_api_init_action()
	{
		register_rest_route( 'mlm-api/v1', '/user/token', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_refresh_token' ),
			'args'		=> array(
				'token'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/register', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_register' ),
			'args'		=> array(
				'uname'		=> array( 'required' => true ),
				'mobile'	=> array( 'required' => true ),
				'pass'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/login', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_login' ),
			'args'		=> array(
				'login'		=> array( 'required' => true ),
				'pass'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/password/lost', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_password_lost' ),
			'args'		=> array(
				'login'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/password/reset', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_password_reset' ),
			'args'		=> array(
				'login'		=> array( 'required' => true ),
				'code'		=> array( 'required' => true ),
				'pass'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/profile/get', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_profile' ),
			'args'		=> array(
				'data'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/profile/save', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_save_profile' ),
			'args'		=> array(
				'fname'		=> array( 'required' => true ),
				'lname'		=> array( 'required' => true ),
				'email'		=> array( 'required' => true ),
				'bio'		=> array( 'required' => false ),
				'avatar'	=> array( 'required' => false ),
				'cover'		=> array( 'required' => false ),
				'mobile'	=> array( 'required' => true ),
				'state'		=> array( 'required' => false ),
				'twitter'	=> array( 'required' => false ),
				'aparat'	=> array( 'required' => false ),
				'telegram'	=> array( 'required' => false ),
				'instagram'	=> array( 'required' => false ),
				'youtube'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/stats', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_user_stats' ),
			'args'		=> array(
				'data'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/upload/file', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_upload_file' ),
			'args'		=> array(
				'file'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/upload/ftp', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_upload_ftp' ),
			'args'		=> array(
				'file'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/get/info', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_site_info' ),
			'args'		=> array(
				'data'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/get/plans', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_plans' ),
			'args'		=> array(
				'data'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/tickets/contacts', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_contacts' ),
			'args'		=> array(
				'data'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/tickets/new', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_new_ticket' ),
			'args'		=> array(
				'recipient'		=> array( 'required' => true ),
				'subject'		=> array( 'required' => true ),
				'title'			=> array( 'required' => true ),
				'content'		=> array( 'required' => true ),
				'unit'			=> array( 'required' => false ),
				'post_id'		=> array( 'required' => false ),
				'attaches'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/tickets/reply', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_reply_ticket' ),
			'args'		=> array(
				'parent'		=> array( 'required' => true ),
				'status'		=> array( 'required' => true ),
				'content'		=> array( 'required' => true ),
				'attaches'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/tickets/search', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_tickets_search' ),
			'args'		=> array(
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/tickets/details', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_ticket_details' ),
			'args'		=> array(
				'ticket_id'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/get/categories', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_categories' ),
			'args'		=> array(
				'id'		=> array( 'required' => false ),
				'type'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/search', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_posts_search' ),
			'args'		=> array(
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
				'category'	=> array( 'required' => false ),
				'keyword'	=> array( 'required' => false ),
				'user_id'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/details', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_posts_details' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/wallet/search', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_wallet_search' ),
			'args'		=> array(
				'type'		=> array( 'required' => true ),
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/wallet/withdrawal', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_withdrawal_request' ),
			'args'		=> array(
				'amount'	=> array( 'required' => true ),
				'card'		=> array( 'required' => true ),
				'sheba'		=> array( 'required' => true ),
				'owner'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/bookmark/search', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_bookmark_search' ),
			'args'		=> array(),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/bookmark/add', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_bookmark_add' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/bookmark/remove', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_bookmark_remove' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/subscribe/plans', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_subscribe_plans' ),
			'args'		=> array(),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/subscribe/new', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_subscribe_new' ),
			'args'		=> array(
				'plan_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/subscribe/history', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_subscribe_history' ),
			'args'		=> array(
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/subscribe/list', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_subscribe_list' ),
			'args'		=> array(
				'plan_id'	=> array( 'required' => true ),
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/access', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_check_access' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/user/follow', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_follow_user' ),
			'args'		=> array(
				'vendor_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/rate', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_rate_post' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
				'rate'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/articles', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_post_articles' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/author', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_post_author' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/comments/get', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_comments_query' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/comments/new', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_new_comment' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
				'name'		=> array( 'required' => true ),
				'email'		=> array( 'required' => true ),
				'comment'	=> array( 'required' => true ),
				'parent'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/me', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_my_posts' ),
			'args'		=> array(
				'search'	=> array( 'required' => false ),
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/get', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_post' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/posts/new', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_new_post' ),
			'args'		=> array(
				'title'		=> array( 'required' => true ),
				'content'	=> array( 'required' => true ),
				'cats'		=> array( 'required' => true ),
				'tags'		=> array( 'required' => true ),
				'post_id'	=> array( 'required' => false ),
				'image_id'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/me', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_my_products' ),
			'args'		=> array(
				'search'	=> array( 'required' => false ),
				'status'	=> array( 'required' => false ),
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/get', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_get_product' ),
			'args'		=> array(
				'post_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/new/file', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_new_file' ),
			'args'		=> array(
				'title'			=> array( 'required' => true ),
				'content'		=> array( 'required' => true ),
				'cats'			=> array( 'required' => true ),
				'tags'			=> array( 'required' => true ),
				'post_id'		=> array( 'required' => false ),
				'image_id'		=> array( 'required' => false ),
				'ref_percent'	=> array( 'required' => true ),
				'price'			=> array( 'required' => true ),
				'sale_price'	=> array( 'required' => false ),
				'button_text'	=> array( 'required' => false ),
				'button_link'	=> array( 'required' => false ),
				'button2_text'	=> array( 'required' => false ),
				'button2_link'	=> array( 'required' => false ),
				'files'			=> array( 'required' => false ),
				'image_thumb'	=> array( 'required' => false ),
				'image_one'		=> array( 'required' => false ),
				'image_two'		=> array( 'required' => false ),
				'fields'		=> array( 'required' => false ),
				'file_type'		=> array( 'required' => false ),
				'page_count'	=> array( 'required' => false ),
				'part_count'	=> array( 'required' => false ),
				'file_author'	=> array( 'required' => false ),
				'file_size'		=> array( 'required' => false ),
				'file_format'	=> array( 'required' => false ),
				'file_language'	=> array( 'required' => false ),
				'file_step'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/new/course', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_new_course' ),
			'args'		=> array(
				'title'			=> array( 'required' => true ),
				'content'		=> array( 'required' => true ),
				'cats'			=> array( 'required' => true ),
				'tags'			=> array( 'required' => true ),
				'post_id'		=> array( 'required' => false ),
				'image_id'		=> array( 'required' => false ),
				'ref_percent'	=> array( 'required' => true ),
				'price'			=> array( 'required' => true ),
				'sale_price'	=> array( 'required' => false ),
				'button_text'	=> array( 'required' => false ),
				'button_link'	=> array( 'required' => false ),
				'button2_text'	=> array( 'required' => false ),
				'button2_link'	=> array( 'required' => false ),
				'image_thumb'	=> array( 'required' => false ),
				'image_one'		=> array( 'required' => false ),
				'image_two'		=> array( 'required' => false ),
				'fields'		=> array( 'required' => false ),
				'file_type'		=> array( 'required' => false ),
				'page_count'	=> array( 'required' => false ),
				'part_count'	=> array( 'required' => false ),
				'file_author'	=> array( 'required' => false ),
				'file_size'		=> array( 'required' => false ),
				'file_format'	=> array( 'required' => false ),
				'file_language'	=> array( 'required' => false ),
				'file_step'		=> array( 'required' => false ),
				'teacher_image'	=> array( 'required' => false ),
				'teacher_name'	=> array( 'required' => false ),
				'course_fill'	=> array( 'required' => false ),
				'teacher_bio'	=> array( 'required' => false ),
				'course_video'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/new/physical', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_new_physical' ),
			'args'		=> array(
				'title'			=> array( 'required' => true ),
				'content'		=> array( 'required' => true ),
				'cats'			=> array( 'required' => true ),
				'tags'			=> array( 'required' => true ),
				'post_id'		=> array( 'required' => false ),
				'image_id'		=> array( 'required' => false ),
				'ref_percent'	=> array( 'required' => true ),
				'price'			=> array( 'required' => true ),
				'sale_price'	=> array( 'required' => false ),
				'button_text'	=> array( 'required' => false ),
				'button_link'	=> array( 'required' => false ),
				'button2_text'	=> array( 'required' => false ),
				'button2_link'	=> array( 'required' => false ),
				'image_thumb'	=> array( 'required' => false ),
				'image_one'		=> array( 'required' => false ),
				'image_two'		=> array( 'required' => false ),
				'fields'		=> array( 'required' => false ),
				'file_type'		=> array( 'required' => false ),
				'page_count'	=> array( 'required' => false ),
				'part_count'	=> array( 'required' => false ),
				'file_author'	=> array( 'required' => false ),
				'file_size'		=> array( 'required' => false ),
				'file_format'	=> array( 'required' => false ),
				'file_language'	=> array( 'required' => false ),
				'file_step'		=> array( 'required' => false ),
				'weight'		=> array( 'required' => false ),
				'quantity'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/course/articles', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_course_articles' ),
			'args'		=> array(
				'course_id'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/course/chapter/save', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_course_chapter_save' ),
			'args'		=> array(
				'course_id'		=> array( 'required' => true ),
				'chapter_id'	=> array( 'required' => false ),
				'image_id'		=> array( 'required' => false ),
				'number'		=> array( 'required' => true ),
				'title'			=> array( 'required' => true ),
				'desc'			=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/course/lesson/save', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_course_lesson_save' ),
			'args'		=> array(
				'course_id'		=> array( 'required' => true ),
				'chapter_id'	=> array( 'required' => true ),
				'lesson_id'		=> array( 'required' => false ),
				'number'		=> array( 'required' => true ),
				'title'			=> array( 'required' => true ),
				'desc'			=> array( 'required' => true ),
				'status'		=> array( 'required' => true ),
				'links'			=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/course/chapter/delete', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_course_chapter_delete' ),
			'args'		=> array(
				'chapter_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/products/course/lesson/delete', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_course_lesson_delete' ),
			'args'		=> array(
				'lesson_id'		=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/upgrade/data', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_upgrade_data' ),
			'args'		=> array(
				'user_id'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/upgrade/step/1', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_step_one' ),
			'args'		=> array(
				'gender'	=> array( 'required' => true ),
				'fname'		=> array( 'required' => true ),
				'lname'		=> array( 'required' => true ),
				'birth'		=> array( 'required' => true ),
				'melli'		=> array( 'required' => true ),
				'address'	=> array( 'required' => true ),
				'phone'		=> array( 'required' => true ),
				'postal'	=> array( 'required' => true ),
				'role'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/upgrade/step/2', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_step_two' ),
			'args'		=> array(
				'melli_file'	=> array( 'required' => true ),
				'shena_file'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/announces/list', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_announces_list' ),
			'args'		=> array(
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
				'status'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/announces/details', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_announces_details' ),
			'args'		=> array(
				'announce_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/links/list', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_links_list' ),
			'args'		=> array(
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
				'mode'		=> array( 'required' => false ),
				'vendor'	=> array( 'required' => false ),
				'search'	=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/referrals/list', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_referrals_list' ),
			'args'		=> array(
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );

		register_rest_route( 'mlm-api/v1', '/subsets/list', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_subsets_list' ),
			'args'		=> array(
				'paged'		=> array( 'required' => false ),
				'count'		=> array( 'required' => false ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );
		
		register_rest_route( 'mlm-api/v1', '/order/attach', array(
			'methods'	=> 'POST',
			'callback'	=> array( $this, 'callback_order_attach' ),
			'args'		=> array(
				'order_id'	=> array( 'required' => true ),
			),
			'permission_callback'	=> function () {
				return true;
			}
		) );
    }


	public function callback_refresh_token( $data )
	{
		$current_id	= $this->find_user_id_by_token();
		$token		= isset( $data['token'] ) ? sanitize_text_field( $data['token'] ) : '';
		$user_id	= $this->get_user_id_from_refresh_token( $token );

		if( ! $current_id )
		{
			return $this->get_error('not_logged_in');
		}

		if( empty( $user_id ) || $current_id != $user_id )
		{
			return $this->get_error('invalid_token');
		}

		if( ! is_user_member_of_blog( $user_id ) )
		{
			return $this->get_error('wrong_blog');
		}

		$userInfo	= $this->refresh_access_token( $token, $user_id );

		if( empty( $userInfo ) )
		{
			return $this->get_error('invalid_token');
		}

		$output	= array(
			'submited'		=> true,
			'access_token'	=> $userInfo['access_token'],
			'expires_in'	=> $userInfo['expires_in']
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $output,
		) );
	}


	public function callback_register( $data )
	{
		$uname		= isset( $data['uname'] ) ? sanitize_text_field( $data['uname'] ) : '';
		$mobile		= isset( $data['mobile'] ) ? sanitize_text_field( $data['mobile'] ) : '';
		$pass		= isset( $data['pass'] ) ? sanitize_text_field( $data['pass'] ) : '';

		if( empty( $uname ) || empty( $pass ) || empty( $mobile ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! preg_match( '/^[a-zA-Z0-9_-]+$/', $uname ) )
		{
			return $this->get_error('invalid_username');
		}

		if( username_exists( $uname ) || strlen( $uname ) < 3 )
		{
			return $this->get_error('username_exists');
		}

		if( ! mlm_is_mobile( $mobile ) )
		{
			return $this->get_error('invalid_mobile');
		}

		if( mlm_mobile_exists( $mobile ) )
		{
			return $this->get_error('mobile_exists');
		}

		if( strlen( $pass ) < 7 )
		{
			return $this->get_error('password_lenght');
		}

		$user_data = array(
			'user_login'	=> $uname,
			'user_pass'		=> $pass,
			'display_name'	=> $uname,
			'first_name'	=> $uname,
			'nickname'		=> $uname,
		);

		$user_id = wp_insert_user( $user_data );

		if( is_wp_error( $user_id ) )
		{
			return $this->get_error('unknown_error');
		}

		update_user_meta( $user_id, 'mlm_mobile', $mobile );

		$output	= array(
			'user_id'		=> $user_id
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $output,
		) );
	}


	public function callback_login( $data )
	{
		$login		= isset( $data['login'] ) ? sanitize_text_field( $data['login'] ) : '';
		$pass		= isset( $data['pass'] ) ? sanitize_text_field( $data['pass'] ) : '';

		if( empty( $login ) || empty( $pass ) )
		{
			return $this->get_error('required_fields');
		}

		$login		= mlm_get_user_by_mobile( $login );
		$userObj	= get_user_by( 'login', $login );

		if( ! $userObj || ! wp_check_password( $pass, $userObj->data->user_pass, $userObj->ID ) )
		{
			return $this->get_error('invalid_login_password');
		}

		// Update access en refresh tokens
		$result	= $this->after_login( $userObj->ID );

		if( $result !== false )
		{
			$output	= array(
				'submited'		=> true,
				'user_id'		=> $userObj->ID,
				'access_token'	=> $result['access_token'],
				'expires_in'	=> $result['expires_in'],
				'refresh_token'	=> $result['refresh_token'],
			);

			return rest_ensure_response( array(
				'status'	=> 13,
				'data'		=> $output,
			) );
		}

		// Failed to generate token
		return rest_ensure_response( $this->get_error('create_token') );
	}


	public function callback_password_lost( $data )
	{
		$login		= isset( $data['login'] ) ? sanitize_text_field( $data['login'] ) : '';

		if( empty( $login ) )
		{
			return $this->get_error('required_fields');
		}

		$user_login	= mlm_get_user_by_mobile( $login );

		if( empty( $user_login ) )
		{
			return $this->get_error('invalid_login');
		}

		if( ! username_exists( $user_login ) && ! email_exists( $user_login ) )
		{
			return $this->get_error('invalid_login');
		}

		if( is_email( $user_login ) && email_exists( $user_login ) )
		{
			$user_obj	= get_user_by( 'email', $user_login );
		}
		else
		{
			$user_obj	= get_user_by( 'login', $user_login );
		}

		$now		= time();
		$db_code	= get_transient( 'mlm_lost_'.$user_obj->ID );
		$db_time	= get_transient( 'mlm_time_'.$user_obj->ID );

		if( ! empty( $db_time ) )
		{
			$diff = $now - absint( $db_time );
		}
		else
		{
			$diff = 200;
		}

		if( $diff < 120 )
		{
			return $this->get_error('wait_for_code');
		}

		$code	= rand( 10000, 99999 );
		$mobile	= get_user_meta( $user_obj->ID, 'mlm_mobile', true );
		$result	= false;

		if( mlm_is_mobile( $mobile ) && mlm_sms_is_active() )
		{
			$result		= mlmFire()->notif->send_user_sms( $user_obj->ID, 'lost_code', array('code' => $code) );
		}

		if( ! $result )
		{
			$result		= mlmFire()->notif->send_user_mail( $user_obj->ID, 'lost_code', array('code' => $code) );
		}

		if( ! $result )
		{
			return $this->get_error('sms_failed');
		}

		set_transient( 'mlm_lost_'.$user_obj->ID , $code, HOUR_IN_SECONDS );
		set_transient( 'mlm_time_'.$user_obj->ID , $now, HOUR_IN_SECONDS );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'code_sent'	=> true
			),
		) );
	}


	public function callback_password_reset( $data )
	{
		$login		= isset( $data['login'] ) ? sanitize_text_field( $data['login'] ) : '';
		$code		= isset( $data['code'] ) ? absint( $data['code'] ) : '';
		$pass		= isset( $data['pass'] ) ? sanitize_text_field( $data['pass'] ) : '';
		$user_login	= mlm_get_user_by_mobile( $login );
		$db_code	= '';

		$user_login	= mlm_get_user_by_mobile( $user_login );

		if( empty( $user_login ) )
		{
			return $this->get_error('invalid_login');
		}

		if( ! username_exists( $user_login ) && ! email_exists( $user_login ) )
		{
			return $this->get_error('invalid_login');
		}

		if( is_email( $user_login ) && email_exists( $user_login ) )
		{
			$user_obj	= get_user_by( 'email', $user_login );
			$db_code	= (int)get_transient( 'mlm_lost_'.$user_obj->ID );
		}
		elseif( username_exists( $user_login ) )
		{
			$user_obj	= get_user_by( 'login', $user_login );
			$db_code	= (int)get_transient( 'mlm_lost_'.$user_obj->ID );
		}

		if( empty( $db_code ) )
		{
			return $this->get_error('no_verify_code');
		}

		if( $code !== $db_code )
		{
			return $this->get_error('code_mismatch');
		}

		if( strlen( $pass ) < 7 )
		{
			return $this->get_error('password_lenght');
		}

		wp_set_password( $pass, $user_obj->ID );
		delete_transient( 'mlm_lost_'.$user_obj->ID );
		delete_transient( 'mlm_time_'.$user_obj->ID );
		delete_transient( 'mlm_reset_'.$user_obj->ID );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'pass_changed'	=> true
			),
		) );
	}


	public function callback_get_profile( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$user_data		= get_userdata( $user_id );
		$role			= 'customer';

		if( user_can($user_id, 'read_private_pages') )
		{
			$role	= 'vendor';
		}
		elseif( user_can($user_id, 'unfiltered_html') )
		{
			$role	= 'referrer';
		}

		$profile		= array(
			'id'		=>	$user_id,
			'login'		=>	$user_data->user_login,
			'fname'		=>	$user_data->first_name,
			'lname'		=>	$user_data->last_name,
			'email'		=>	$user_data->user_email,
			'bio'		=>	$user_data->description,
			'avatar'	=>	get_user_meta( $user_id, 'mlm_avatar', true ),
			'cover'		=>	get_user_meta( $user_id, 'mlm_cover', true ),
			'mobile'	=>	get_user_meta( $user_id, 'mlm_mobile', true ),
			'state'		=>	get_user_meta( $user_id, 'mlm_state', true ),
			'twitter'	=>	get_user_meta( $user_id, 'mlm_twitter', true ),
			'aparat'	=>	get_user_meta( $user_id, 'mlm_aparat', true ),
			'telegram'	=>	get_user_meta( $user_id, 'mlm_telegram', true ),
			'instagram'	=>	get_user_meta( $user_id, 'mlm_instagram', true ),
			'youtube'	=>	get_user_meta( $user_id, 'mlm_youtube', true ),
			'role'		=>	$role
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $profile,
		) );
	}


	public function callback_save_profile( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$user_data	= get_userdata( $user_id );
		$db_email	= $user_data->user_email;
		$fname		= isset( $data['fname'] ) ? sanitize_text_field( $data['fname'] ) : '';
		$lname		= isset( $data['lname'] ) ? sanitize_text_field( $data['lname'] ) : '';
		$email		= isset( $data['email'] ) ? sanitize_text_field( $data['email'] ) : '';
		$bio		= isset( $data['bio'] ) ? esc_textarea( $data['bio'] ) : '';
		$avatar		= isset( $data['avatar'] ) ? esc_url( $data['avatar'] ) : '';
		$cover		= isset( $data['cover'] ) ? esc_url( $data['cover'] ) : '';
		$mobile		= isset( $data['mobile'] ) ? sanitize_text_field( $data['mobile'] ) : '';
		$state		= isset( $data['state'] ) ? sanitize_text_field( $data['state'] ) : '';
		$twitter	= isset( $data['twitter'] ) ? esc_url( $data['twitter'] ) : '';
		$aparat		= isset( $data['aparat'] ) ? esc_url( $data['aparat'] ) : '';
		$telegram	= isset( $data['telegram'] ) ? esc_url( $data['telegram'] ) : '';
		$instagram	= isset( $data['instagram'] ) ? esc_url( $data['instagram'] ) : '';
		$youtube	= isset( $data['youtube'] ) ? esc_url( $data['youtube'] ) : '';

		if( empty( $fname ) || empty( $lname ) || empty( $email ) || empty( $mobile ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! is_email( $email ) )
		{
			return $this->get_error('invalid_email');
		}

		if( email_exists( $email ) && $db_email != $email )
		{
			return $this->get_error('email_exists');
		}

		if( ! mlm_is_mobile( $mobile ) )
		{
			return $this->get_error('invalid_mobile');
		}

		if( mlm_mobile_exists( $mobile, $user_id ) )
		{
			return $this->get_error('mobile_exists');
		}

		$userdata = array(
			'ID' 			=> $user_id,
			'first_name'	=> $fname,
			'last_name'		=> $lname,
			'user_email'	=> $email,
			'nickname'		=> $fname .' '. $lname,
			'display_name'	=> $fname .' '. $lname,
			'description'	=> $bio,
		);
		wp_update_user( $userdata );

		update_user_meta( $user_id, 'mlm_avatar', $avatar );
		update_user_meta( $user_id, 'mlm_cover', $cover );
		update_user_meta( $user_id, 'mlm_mobile', $mobile );
		update_user_meta( $user_id, 'mlm_state', $state );
		update_user_meta( $user_id, 'mlm_twitter', $twitter );
		update_user_meta( $user_id, 'mlm_aparat', $aparat );
		update_user_meta( $user_id, 'mlm_telegram', $telegram );
		update_user_meta( $user_id, 'mlm_instagram', $instagram );
		update_user_meta( $user_id, 'mlm_youtube', $youtube );

		mlmFire()->dashboard->get_profile_status( $user_id, true );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'user_id'	=> $user_id,
			),
		) );
	}


	public function callback_get_user_stats( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$chart_data	= mlmFire()->dashboard->get_user_chart_data( $user_id );
		$output		= array(
			'verify_status'		=> mlmFire()->dashboard->get_account_status( $user_id ),
			'profile_fill'		=> mlmFire()->dashboard->get_profile_status( $user_id ),
			'chart_data'		=> ( is_array( $chart_data ) ) ? array_values( $chart_data ) : array(),
			'balance'			=> mlmFire()->wallet->get_balance( $user_id ),
			'total_withdrawal'	=> mlmFire()->wallet->total_withdraw_amount( $user_id ),
			'last_withdrawal'	=> mlmFire()->wallet->recent_withdraw_amount( $user_id ),
			'total_income'		=> mlmFire()->wallet->total_income_amount( $user_id ),
			'subsets_cnt'		=> mlmFire()->network->get_subs_count( $user_id ),
			'sales_cnt'			=> mlmFire()->wallet->get_user_sales_count( $user_id ),
			'posts_cnt'			=> count_user_posts( $user_id , 'post' ),
			'products_cnt'		=> count_user_posts( $user_id , 'product' ),
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $output,
		) );
	}


	public function callback_upload_file( WP_REST_Request $request )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$files	= isset( $_FILES['file'] ) ? $_FILES['file'] : '';

		if( empty( $files ) )
		{
			return $this->get_error('no_file_selected');
		}

		if( $files['error'] === 0 )
		{
			$allowed	= array( 'jpg', 'jpeg', 'png', 'gif', 'pdf', 'doc', 'docx', 'zip' );

			$temp = $files['tmp_name'];
			$ext  = explode( '.', $files['name'] );
			$ext  = strtolower( end( $ext ) );
			$file = uniqid( '', true ) . time() . '.' . $ext;

			$uploads	= wp_upload_dir();
			$path		= str_replace('\\', '/', $uploads['path']);

			if( ! file_exists( $path ) )
			{
				wp_mkdir_p( $path );
			}

			if( ! in_array( $ext, $allowed ) )
			{
				return $this->get_error('file_wrong_format');
			}

			if( $files['size'] > 5242880 ) // 5 MB
			{
				return $this->get_error('file_too_large');
			}

			if( move_uploaded_file($temp, $path .'/'. $file) )
			{
				$filetype = wp_check_filetype( basename( $path . '/' . $file ), null );

				$attachment = array(
					'guid'				=> $uploads['url'] .'/'. $file,
					'post_mime_type'	=> $filetype['type'],
					'post_title'		=> $file,
					'post_content'		=> '',
					'post_status'		=> 'inherit'
				);
				$attach_id	= wp_insert_attachment( $attachment, $path . '/' . $file, 0 );

				require_once( ABSPATH . 'wp-admin/includes/image.php' );
				$attach_data	= wp_generate_attachment_metadata( $attach_id, $path . '/' . $file );
				wp_update_attachment_metadata( $attach_id, $attach_data );

				$output	= array(
					'file_id'	=> $attach_id,
					'file_url'	=> $uploads['url'] .'/'. $file,
				);

				return rest_ensure_response( array(
					'status'	=> 13,
					'data'		=> $output,
				) );
			}
		}

		return $this->get_error('unknown_error');
	}


	public function callback_upload_ftp( WP_REST_Request $request )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		if( ! user_can( $user_id, 'upload_files') )
        {
            return $this->get_error('not_allowed');
        }

        $user_data	= get_userdata($user_id);
        $user_name	= $user_data->user_login;
        $max		= 2122317824; // 2GB
        $ftp_url	= get_option('mlm_ftp_url');
        $ftp_user	= get_option('mlm_ftp_user');
        $ftp_pass	= get_option('mlm_ftp_pass');
        $ftp_link	= get_option('mlm_ftp_link');
		$files		= isset( $_FILES['file'] ) ? $_FILES['file'] : '';

		if( empty( $ftp_url ) || empty( $ftp_user ) || empty( $ftp_pass ) || empty( $ftp_link ) )
        {
            return $this->get_error('unknown_error');
        }

		if( empty( $files ) )
		{
			return $this->get_error('no_file_selected');
		}

		if( $files['error'] === 0 )
		{
			$uploads	= wp_upload_dir();
	        $allowed	= array( 'mp3', 'mp4', 'png', 'jpg', 'gif', 'mpg', 'amr', 'avi', 'wma', 'm4a', 'aac', 'wav', 'pdf', 'ppt', 'doc', 'docx', 'text', 'excel', 'zip', 'epub', 'srt', 'mkv' );
			$path		= str_replace( '\\', '/', $uploads['basedir'] );
			$base_url	= str_replace( '\\', '/', $uploads['baseurl'] );
			$sub		= str_replace( '\\', '/', $uploads['subdir'] );
			$folder		= trailingslashit( $path ) . $sub;
			$folder_url	= trailingslashit( $base_url ) . $sub;

			if( ! file_exists( $folder ) )
			{
				wp_mkdir_p( $folder );
				/* mkdir( $folder, 0777, true ); */
			}

			$temp	= $file['tmp_name'];
			$ext	= explode( '.', $file['name'] );
			$ext	= strtolower( end( $ext ) );
			$full	= $user_name . '-' . time() . '.' . $ext;

			if( ! in_array( $ext, $allowed ) )
			{
				return $this->get_error('file_wrong_format');
			}

			if( $file['size'] > $max )
			{
				return $this->get_error('file_too_large');
			}

			if( move_uploaded_file( $temp, $folder .'/'. $full ) )
			{
				$connection = ftp_connect( $ftp_url );

				if( @ftp_login($connection, $ftp_user, $ftp_pass) )
				{
					ftp_put($connection, $full, $folder . '/' . $full, FTP_BINARY);
					ftp_close($connection);

					return rest_ensure_response( array(
						'status'	=> 13,
						'data'		=> array(
							'url'		=> trailingslashit( $ftp_link ) . $full,
							'uploaded'	=> true,
						),
					) );
				}

				@unlink( $folder . '/' . $full );
			}
		}

		return $this->get_error('unknown_error');
	}


	public function callback_get_site_info( $data )
	{
		// SLIDER
		$slides	= array();
		for( $i = 1; $i <= 8; $i++ )
		{
			$slide_txt	= get_option( 'mlm_slide_txt_' . $i );
			$slide_url	= esc_url( get_option( 'mlm_slide_url_' . $i ) );
			$slide_img	= esc_url( get_option( 'mlm_slide_img_' . $i ) );

			if( empty( $slide_url ) )
			{
				$slide_url = '#';
			}

			if( ! empty( $slide_img ) && ! empty( $slide_txt ) )
			{
				$slides[] = array(
					'txt'	=> $slide_txt,
					'url'	=> $slide_url,
					'img'	=> $slide_img,
				);
			}
		}

		// STATS
		$all_posts		= wp_count_posts('post');
		$all_products	= wp_count_posts('product');
		$all_users		= count_users();
		$all_comments	= wp_count_comments();

		// OUTPUT
		$output	= array(
			'slides'			=> $slides,
			'top_vendors'		=> mlmFire()->dashboard->get_top_vendors(),
			'top_referrers'		=> mlmFire()->dashboard->get_top_referrers(),
			'app_color'			=> get_option('mlm_app_color'),
			'app_banner'		=> get_option('mlm_app_banner'),
			'app_product_cat'	=> get_option('mlm_app_product_cat'),
			'app_blog_cat'		=> get_option('mlm_app_blog_cat'),
			'app_stats_box'		=> get_option('mlm_app_stats_box'),
			'app_top_banner'	=> get_option('mlm_app_top_banner'),
			'app_upgrade_ref'	=> get_option('mlm_app_upgrade_ref'),
			'app_upgrade_sell'	=> get_option('mlm_app_upgrade_sell'),
			'app_rtl'			=> get_option('mlm_app_rtl'),
			'stats'				=> array(
				'posts_cnt'			=> isset( $all_posts->publish ) ? $all_posts->publish : 0,
				'products_cnt'		=> isset( $all_products->publish ) ? $all_products->publish : 0,
				'users_cnt'			=> isset( $all_users['total_users'] ) ? $all_users['total_users'] : 0,
				'comments_cnt'		=> isset( $all_comments->approved ) ? $all_comments->approved : 0,
				'tickets_cnt'		=> mlmFire()->ticket->count_all_tickets(),
				'sales_cnt'			=> mlmFire()->wallet->get_total_sales_count(),
			)
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $output,
		) );
	}


	public function callback_get_plans( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$all_plans		= mlmFire()->plan->get_plans( 0, true );
		$current_plans	= mlmFire()->plan->get_user_active_plan( $user_id );

		if( ! is_array( $all_plans ) || ! count( $all_plans ) )
		{
			$all_plans	= array();
		}
		else
		{
			$all_plans	= array_values( $all_plans );
		}

		if( ! is_array( $current_plans ) || ! count( $current_plans ) )
		{
			$current_plans	= array();
		}
		else
		{
			$current_plans	= array_values( $current_plans );
		}

		// OUTPUT
		$output	= array(
			'plans'			=> $all_plans,
			'active_plans'	=> $current_plans
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $output,
		) );
	}


	public function callback_get_contacts( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$contacts	= array( 0 => array(
			'vendor'	=> 0,
			'title'		=> 'پشتیبانی سایت'
		) );
		$purchases	= get_posts( array(
			'numberposts' => -1,
			'meta_key'    => '_customer_user',
			'meta_value'  => $user_id,
			'post_type'   => wc_get_order_types(),
			'post_status' => array_keys( wc_get_is_paid_statuses() ),
		) );

		if( $purchases )
		{
			foreach( $purchases as $payment )
			{
				$order	= wc_get_order( $payment->ID );
				$items	= $order->get_items();

				foreach( $items as $item )
				{
					$product_id = $item->get_product_id();
					$vendor_id	= get_post_field( 'post_author', $product_id );

					if( ! isset( $contacts[$product_id] ) && mlm_user_exists( $vendor_id ) )
					{
						$contacts[$product_id]	= array(
							'post_id'	=> $product_id,
							'vendor'	=> $vendor_id,
							'title'		=> get_the_title( $product_id ),
						);
					}
				}
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array_values( $contacts ),
		) );
	}


	public function callback_new_ticket( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$recipient	= isset( $data['recipient'] ) ? absint( $data['recipient'] ) : 0;
		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
		$subject	= isset( $data['subject'] ) ? absint( $data['subject'] ) : '';
		$unit		= isset( $data['unit'] ) ? sanitize_text_field( $data['unit'] ) : '';
		$title		= isset( $data['title'] ) ? sanitize_text_field( $data['title'] ) : '';
		$content	= isset( $data['content'] ) ? wp_filter_post_kses( $data['content'] ) : '';
		$attaches	= isset( $data['attaches'] ) ? mlm_sanitize_array( $data['attaches'] ) : array();

		if( empty( $subject ) )
		{
			return $this->get_error('subject_required');
		}

		if( $subject == 1 && ! mlm_user_exists( $recipient ) )
		{
			return $this->get_error('related_product_required');
		}
		elseif( $subject == 2 && empty( $unit ) )
		{
			return $this->get_error('related_department_required');
		}

		if( empty( $title ) || empty( $content ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! empty( $recipient ) && ! mlm_user_exists( $recipient ) )
		{
			return $this->get_error('invalid_contact_id');
		}

		if( ! empty( $user_id ) && $recipient === $user_id )
		{
			return $this->get_error('no_ticket_for_yourself');
		}

		if( ! mlm_user_exists( $recipient ) && ! empty( $unit ) )
		{
			$user_data	= array( 'unit'	=> $unit );
		}
		else
		{
			$user_data	= array();
		}

		$ticket_id = mlmfire()->db->ticket_record( $post_id, $user_id, $recipient, $title, $content, 0, $user_data, 1, $attaches );

		if( ! $ticket_id )
		{
			return $this->get_error('unknown_error');
		}

		// NOTIFICATION
		mlmFire()->notif->send_user_mail( $user_id, 'new_ticket', array(
			'ticket_id'	=> $ticket_id
		) );

		if( ! mlm_user_exists( $recipient ) )
		{
			mlmFire()->notif->send_admin_sms( 'new_ticket', array(
				'ticket_id'	=> $ticket_id
			) );
		}
		else
		{
			mlmFire()->notif->send_user_sms( $recipient, 'new_ticket', array(
				'ticket_id'	=> $ticket_id
			) );
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'ticket_id'	=> $ticket_id
			),
		) );
	}


	public function callback_reply_ticket( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$parent		= isset( $data['parent'] ) ? absint( $data['parent'] ) : 0;
		$status		= isset( $data['status'] ) ? absint( $data['status'] ) : 0;
		$content	= isset( $data['content'] ) ? wp_filter_post_kses( $data['content'] ) : '';
		$attaches	= isset( $data['attaches'] ) ? mlm_sanitize_array( $data['attaches'] ) : array();
		$ticket_obj	= mlmfire()->ticket->get_ticket_data( $parent );

		if( empty( $status ) || empty( $content ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! isset( $ticket_obj->sender_id ) )
		{
			return $this->get_error('invalid_ticket_id');
		}

		if( ! user_can( $user_id, 'manage_options' ) && ( $user_id != $ticket_obj->sender_id && $user_id != $ticket_obj->reciver_id ) )
		{
			return $this->get_error('not_allowed');
		}

		$ticket_id	= mlmfire()->db->ticket_record( 0, $user_id, 0, 'پاسخ', $content, $parent, aaray(), 1, $attaches );

		if( ! $ticket_id )
		{
			return $this->get_error('unknown_error');
		}

		mlmfire()->db->ticket_update( $parent, array( 'status' => $status ) );

		// NOTIFICATION
		if( $user_id == $ticket_obj->sender_id )
		{
			if( ! mlm_user_exists( $ticket_obj->reciver_id ) )
			{
				mlmFire()->notif->send_admin_sms( 'new_ticket', array(
					'ticket_id'	=> $ticket_id
				) );
			}
			else
			{
				mlmFire()->notif->send_user_sms( $ticket_obj->reciver_id, 'ticket_replied', array(
					'ticket_id'	=> $ticket_id
				) );
			}
		}
		elseif( $user_id == $ticket_obj->reciver_id )
		{
			mlmFire()->notif->send_user_sms( $ticket_obj->sender_id, 'ticket_replied', array(
				'ticket_id'	=> $ticket_id
			) );
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'ticket_id'	=> $ticket_id,
			),
		) );
	}


	public function callback_tickets_search( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$list		= array();
		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$count		= isset( $data['count'] ) ? absint( $data['count'] ) : 15;
		$start		= intval( ($paged - 1) * $count );
		$results	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE parent_id = %d AND ( sender_id = %d OR reciver_id = %d ) ORDER BY id DESC LIMIT %d, %d",
			array( 0, $user_id, $user_id, $start, $count ),
			'ticket'
		);
		$all_pages	= mlmFire()->db->count_query_rows(
			"SELECT COUNT(id) FROM {TABLE} WHERE parent_id = %d AND ( sender_id = %d OR reciver_id = %d )",
			array( 0, $user_id, $user_id ),
			'ticket'
		);

		if( ! empty( $results ) )
		{
			foreach( $results as $item )
			{
				$updated	= mlmFire()->ticket->get_last_change( $item->id, $item->date );
				$list[]		= array(
					'id'			=> $item->id,
					'title'			=> $item->title,
					'sender'		=> mlm_get_user_name( $item->sender_id, 'پشتیبانی سایت' ),
					'reciver'		=> mlm_get_user_name( $item->reciver_id, 'پشتیبانی سایت' ),
					'status'		=> $item->status,
					'status_fa'		=> strip_tags( mlmFire()->ticket->ticket_status( $item->status ) ),
					'date'			=> $item->date,
					'local_date'	=> date_i18n( 'Y/m/d H:i:s', strtotime( $item->date ) ),
					'updated'		=> human_time_diff( strtotime( $updated ), current_time('timestamp') ) . ' قبل',
				);
			}
		}

		return rest_ensure_response( array(
			'status'		=> 13,
			'crypt_data'	=> array(
				'count'			=> count( $list ),
				'all_pages'		=> $all_pages,
				'items'			=> $list
			),
		) );
	}


	public function callback_ticket_details( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$list		= array();
		$ticket_id	= isset( $data['ticket_id'] ) ? absint( $data['ticket_id'] ) : '';

		if( empty( $ticket_id ) )
		{
			return $this->get_error('invalid_ticket_id');
		}

		$ticket_obj	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE id = %d AND parent_id = %d AND ( sender_id = %d OR reciver_id = %d ) LIMIT %d",
			array( $ticket_id, 0, $user_id, $user_id, 1 ),
			'ticket',
			true
		);

		if( empty( $ticket_obj ) )
		{
			return $this->get_error('invalid_ticket_id');
		}

		$updated	= mlmFire()->ticket->get_last_change( $ticket_obj->id, $ticket_obj->date );
		$reply		= array();
		$replies	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE parent_id = %d ORDER BY id ASC",
			array( $ticket_id ),
			'ticket'
		);

		if( ! empty( $replies ) )
		{
			foreach( $replies as $v )
			{
				$reply[] = array(
					'sender'		=> mlm_get_user_name( $v->sender_id, 'پشتیبانی سایت' ),
					'content'		=> $v->content,
					'date'			=> $v->date,
					'local_date'	=> date_i18n( 'Y/m/d H:i:s', strtotime( $v->date ) ),
					'attaches'		=> isset( $v->attaches ) ? maybe_unserialize( $v->attaches ) : array(),
				);
			}
		}

		$ticket	= array(
			'id'			=> $ticket_obj->id,
			'title'			=> $ticket_obj->title,
			'post_id'		=> $ticket_obj->post_id,
			'post_title'	=> mlm_get_post_title( $ticket_obj->post_id, '' ),
			'content'		=> $ticket_obj->content,
			'sender'		=> mlm_get_user_name( $ticket_obj->sender_id, 'پشتیبانی سایت' ),
			'reciver'		=> mlm_get_user_name( $ticket_obj->reciver_id, 'پشتیبانی سایت' ),
			'status'		=> $ticket_obj->status,
			'status_fa'		=> strip_tags( mlmFire()->ticket->ticket_status( $ticket_obj->status ) ),
			'date'			=> $ticket_obj->date,
			'local_date'	=> date_i18n( 'Y/m/d H:i:s', strtotime( $ticket_obj->date ) ),
			'updated'		=> human_time_diff( strtotime( $updated ), current_time('timestamp') ) . ' قبل',
			'attaches'		=> isset( $ticket_obj->attaches ) ? maybe_unserialize( $ticket_obj->attaches ) : array(),
			'replies'		=> $reply,
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $ticket,
		) );
	}


	public function callback_get_categories( $data )
	{
		$id		= isset( $data['id'] ) ? absint( $data['id'] ) : '';
		$type	= isset( $data['type'] ) ? sanitize_text_field( $data['type'] ) : '';

		switch( $type )
		{
			case 'tag':
				$type = 'post_tag';
				break;

			case 'product_category':
				$type = 'product_cat';
				break;

			case 'product_tag':
				$type = 'product_tag';
				break;

			default:
				$type = 'category';
				break;
		}

		if( ! empty( $id ) )
		{
			$category	= get_term( $id );

			if( empty( $category ) || is_wp_error( $category ) )
			{
				return $this->get_error('no_items_found');
			}

			$childs	= array();
			$level2	= get_terms( array(
				'taxonomy'		=> $type,
				'hide_empty'	=> 0,
				'parent'		=> $id,
			) );

			if( ! empty( $level2 ) && ! is_wp_error( $level2 ) )
			{
				foreach( $level2 as $child )
				{
					$childs[]			= array(
						'id'			=> $child->term_id,
						'name'			=> $child->name,
						'description'	=> $child->description,
						'count'			=> $child->count
					);
				}
			}

			$list	= array(
				'id'			=> $id,
				'name'			=> $category->name,
				'description'	=> $category->description,
				'count'			=> $category->count,
				'childs'		=> $childs
			);
		}
		else
		{
			$category_terms	= get_terms( array(
				'taxonomy'		=> $type,
				'hide_empty'	=> 0,
				'parent'		=> 0,
			) );

			if( empty( $category_terms ) || is_wp_error( $category_terms ) )
			{
				return $this->get_error('no_items_found');
			}

			$list	= array();

			foreach( $category_terms as $category )
			{
				$obj	= array(
					'id'			=> $category->term_id,
					'name'			=> $category->name,
					'description'	=> $category->description,
					'count'			=> $category->count
				);
				$childs	= array();
				$level2	= get_terms( array(
					'taxonomy'		=> $type,
					'hide_empty'	=> 0,
					'parent'		=> $category->term_id,
				) );

				if( ! empty( $level2 ) && ! is_wp_error( $level2 ) )
				{
					foreach( $level2 as $child )
					{
						$ch_obj	= array(
							'id'			=> $child->term_id,
							'name'			=> $child->name,
							'description'	=> $child->description,
							'count'			=> $child->count
						);
						$subs	= array();
						$level3	= get_terms( array(
							'taxonomy'		=> $type,
							'hide_empty'	=> 0,
							'parent'		=> $child->term_id,
						) );

						if( ! empty( $level3 ) && ! is_wp_error( $level3 ) )
						{
							foreach( $level3 as $sub )
							{
								$subs[]	= array(
									'id'			=> $sub->term_id,
									'name'			=> $sub->name,
									'description'	=> $sub->description,
									'count'			=> $sub->count,
									'childs'		=> array(),
								);
							}
						}

						$ch_obj['childs']	= $subs;
						$childs[]			= $ch_obj;
					}
				}

				$obj['childs']	= $childs;
				$list[]			= $obj;
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $list,
		) );
	}


	public function callback_posts_search( $data )
	{
		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$count		= isset( $data['count'] ) ? absint( $data['count'] ) : 20;
		$keyword	= isset( $data['keyword'] ) ? sanitize_text_field( $data['keyword'] ) : '';
		$category	= isset( $data['category'] ) ? absint( $data['category'] ) : '';
		$user_id	= isset( $data['user_id'] ) ? absint( $data['user_id'] ) : '';

		if( $count > 20 )
		{
			$count = 20;
		}

		$args		= array(
			'post_type'				=> 'post',
			'post_status'			=> 'publish',
			'ignore_sticky_posts'	=> 1,
			'paged'					=> $paged,
			'posts_per_page'		=> $count,
			's'						=> $keyword,
		);

		if( ! empty( $user_id ) )
		{
			$args['author']	= $user_id;
		}

		if( ! empty( $category ) )
		{
			$args['tax_query']	= array( array(
				'taxonomy'	=> 'category',
				'field'		=> 'term_id',
				'terms'		=> array( $category ),
			) );
		}

		$list		= array();
		$p_query	= new WP_Query( $args );

		if( $p_query->have_posts() )
		{
			while( $p_query->have_posts() )
			{
				$p_query->the_post();

				$post_id	= get_the_ID();
				$cats		= array();
				$categories	= wp_get_post_terms( $post_id, 'category' );

				if( ! empty( $categories ) && ! is_wp_error( $categories ) )
				{
					foreach( $categories as $cat )
					{
						$cats[] = array(
							'id'	=> $cat->term_id,
							'name'	=> $cat->name
						);
					}
				}

				$list[]		= array(
					'id'			=> $post_id,
					'title'			=> get_the_title(),
					'excerpt'		=> mlm_excerpt( 250, false ),
					'cats'			=> $cats,
					'image'			=> get_the_post_thumbnail_url( $post_id, 'medium' ),
					'views'			=> mlm_get_post_views( $post_id ),
					'date'			=> get_the_date( 'Y-m-d' ),
					'time'			=> get_the_date( 'H:i:s' ),
					'url'			=> wp_get_shortlink(),
				);
			}

			wp_reset_postdata();
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'count'		=> count( $list ),
				'all_pages'	=> isset( $p_query->max_num_pages ) ? absint( $p_query->max_num_pages ) : 0,
				'items'		=> $list,
			),
		) );
	}


	public function callback_posts_details( $data )
	{
		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : '';
		$post_type	= get_post_type( $post_id );

		if( ! mlm_post_exists( $post_id ) || $post_type != 'post' )
		{
			return $this->get_error('invalid_post_id');
		}

		mlm_set_post_views( $post_id );

		$tags		= array();
		$category	= array();
		$categories	= wp_get_post_terms( $post_id, 'category' );
		$tags_arr	= wp_get_post_terms( $post_id, 'post_tag' );

		if( ! empty( $categories ) && ! is_wp_error( $categories ) )
		{
			foreach( $categories as $cat )
			{
				$category[] = array(
					'id'	=> $cat->term_id,
					'name'	=> $cat->name
				);
			}
		}

		if( ! empty( $tags_arr ) && ! is_wp_error( $tags_arr ) )
		{
			foreach( $tags_arr as $tg )
			{
				$tags[] = array(
					'id'	=> $tg->term_id,
					'name'	=> $tg->name
				);
			}
		}

		$author_id	= get_post_field( 'post_author', $post_id );
		$post		= array(
			'title'			=> get_the_title( $post_id ),
			'description'	=> get_post_field( 'post_content', $post_id ),
			'author_id'		=> $author_id,
			'author_name'	=> get_the_author_meta( 'display_name', $author_id ),
			'image'			=> get_the_post_thumbnail_url( $post_id, 'large' ),
			'category'		=> $category,
			'tags'			=> $tags,
			'views'			=> mlm_get_post_views( $post_id ),
			'date'			=> get_the_date( 'Y-m-d', $post_id ),
			'time'			=> get_the_date( 'H:i:s', $post_id ),
			'url'			=> wp_get_shortlink( $post_id ),
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $post,
		) );
	}


	public function callback_wallet_search( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$type		= isset( $data['type'] ) ? absint( $data['type'] ) : 1;
		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$count		= isset( $data['count'] ) ? absint( $data['count'] ) : 20;
		$start		= intval( ($paged - 1) * $count );
		$results	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE user_id = %d AND type = %d ORDER BY id DESC LIMIT %d, %d",
			array( $user_id, $type, $start, $count ),
			'wallet'
		);
		$all_pages	= mlmFire()->db->count_query_rows(
			"SELECT COUNT(id) FROM {TABLE} WHERE user_id = %d AND type = %d",
			array( $user_id, $type ),
			'wallet'
		);

		$transactions	= array();
		if( ! empty( $results ) )
		{
			foreach( $results as $result )
			{
				$transactions[]	= array(
					'id'		=> $result->id,
					'amount'	=> $result->amount,
					'type'		=> $result->type,
					'type_fa'	=> mlmFire()->wallet->get_type_text( $result->type ),
					'status'	=> mlmFire()->wallet->get_status_text( $result->status ),
					'track'		=> strtotime( $result->date ),
					'text'		=> $result->description,
					'date'		=> date_i18n( 'j F Y', strtotime( $result->date ) ),
					'time'		=> date_i18n( 'H:i', strtotime( $result->date ) ),
				);
			}
		}

		$output	= array(
			'balance'			=> mlmFire()->wallet->get_balance( $user_id ),
			'min_withdrawal'	=> mlmFire()->wallet->min_withdraw_amount(),
			'all_pages'			=> $all_pages,
			'count'				=> count( $transactions ),
			'items'				=> $transactions,
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $output,
		) );
	}


	public function callback_withdrawal_request( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$amount		= isset( $data['amount'] ) ? absint( $data['amount'] ) : 0;
		$card		= isset( $data['card'] ) ? sanitize_text_field( $data['card'] ) : '';
		$sheba		= isset( $data['sheba'] ) ? sanitize_text_field( $data['sheba'] ) : '';
		$owner		= isset( $data['owner'] ) ? sanitize_text_field( $data['owner'] ) : '';
		$balance	= mlmFire()->wallet->get_balance( $user_id );
		$min		= mlmFire()->wallet->min_withdraw_amount();
		$pending	= mlmFire()->db->query_rows(
			"SELECT id FROM {TABLE} WHERE user_id = %d AND type = %d AND status = %d LIMIT %d",
			array( $user_id, 5, 1, 1 ),
			'wallet',
			true
		);

		if( empty( $amount ) || empty( $card ) || empty( $sheba ) || empty( $owner ) )
		{
			return $this->get_error('required_fields');
		}

		if( $amount > $balance )
		{
			return $this->get_error('insufficient_funds');
		}

		if( $amount < $min || $amount <= 0 )
		{
			return $this->get_error('amount_less_than_minimum');
		}

		if( ! empty( $pending ) )
		{
			return $this->get_error('pending_withdrawal');
		}

		$id = mlmFire()->db->wallet_record(
			$user_id,
			0,
			0,
			$amount,
			5,
			1,
			'در انتظار بررسی و پرداخت'
		);
		mlmFire()->wallet->update_meta( $user_id, 'mlm_balance', $amount, 'minus' );

		// SAVE ACCOUNT
		update_user_meta( $user_id, 'mlm_card', $card );
		update_user_meta( $user_id, 'mlm_sheba', $sheba );
		update_user_meta( $user_id, 'mlm_owner', $owner );

		// NOTIFICATION
		mlmFire()->notif->send_admin_mail( 'withdrawal', array(
			'user_id'	=> $user_id,
			'amount'	=> $amount
		) );

		mlmFire()->notif->send_user_mail( $user_id, 'withdrawal_request', array(
			'amount'	=> $amount
		) );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'id'		=> $id,
			),
		) );
	}


	public function callback_bookmark_search( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$bookmarks		= mlmFire()->rating->get_bookmarks( $user_id );
		$post_books		= array();
		$product_books	= array();

		foreach( (array)$bookmarks as $key => $value )
		{
			if( $value == 'product' )
			{
				$product_books[] = array(
					'post_id'	=> $key,
					'title'		=> mlm_get_post_title( $key ),
					'image'		=> mlm_image_url( $key, 'thumbnail', false ),
					'price'		=> mlm_filter( mlm_get_product_price( $key ) ),
				);
			}
			elseif( $value == 'post' )
			{
				$post_books[] = array(
					'post_id'	=> $key,
					'title'		=> mlm_get_post_title( $key ),
					'image'		=> mlm_image_url( $key, 'thumbnail', false )
				);
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'posts'		=> $post_books,
				'products'	=> $product_books,
			),
		) );
	}


	public function callback_bookmark_add( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
		$result		= mlmFire()->rating->bookmark_post( $user_id, $post_id );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'post_id'	=> $post_id
			),
		) );
	}


	public function callback_bookmark_remove( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
		$result		= mlmFire()->rating->remove_post_bookmark( $user_id, $post_id );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'post_id'	=> $post_id
			),
		) );
	}


	public function callback_subscribe_plans( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$all_plans		= mlmFire()->plan->get_plans( 0, true );
		$current_plans	= mlmFire()->plan->get_user_active_plan( $user_id );
		$plans_obj		= array();

		if( is_array( $all_plans ) && count( $all_plans ) > 0 )
		{
			foreach( $all_plans as $plan )
			{
				$plans_obj[] = array(
					'id'		=> $plan['id'],
					'name'		=> $plan['name'],
					'text'		=> $plan['text'],
					'time'		=> $plan['time'],
					'price'		=> mlm_filter( $plan['price'] ),
					'active'	=> ( is_array( $current_plans ) && in_array( $plan['id'], $current_plans ) ) ? true : false,
				);
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'plans'		=> $plans_obj
			),
		) );
	}


	public function callback_subscribe_new( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$plan_id		= isset( $data['plan_id'] ) ? absint( $data['plan_id'] ) : 0;
		$plan_data		= mlmFire()->plan->get_plans( $plan_id );
		$user_balance	= mlmFire()->wallet->get_balance( $user_id );

		if( empty( $plan_id ) || ! isset( $plan_data['id'] ) )
		{
			return $this->get_error('invalid_plan_id');
		}

		if( $plan_data['time'] < 1 )
		{
			return $this->get_error('invalid_plan_time');
		}

		if( $user_balance < $plan_data['price'] )
		{
			return $this->get_error('balance_less_than_amount');
		}

		$order_id	= mlmFire()->db->wallet_record(
			$user_id,
			0,
			0,
			$plan_data['price'],
			8,
			2,
			'خرید پلن اشتراک ویژه با موجودی کیف پول'
		);

		if( ! $order_id )
		{
			return $this->get_error('unknown_error');
		}

		// DECREASE BALANCE
		mlmFire()->wallet->update_meta( $user_id, 'mlm_balance', $plan_data['price'], 'minus' );

		// SAVE SUBSCRIBE
		$sub_id	= mlmFire()->db->subscribe_record( $user_id, $order_id, $plan_data );

		if( ! $sub_id )
		{
			return $this->get_error('unknown_error');
		}

		mlmFire()->db->subscribe_update( $sub_id, array( 'status' => 1 ) );

		// ACTIVATE PLAN
		mlmFire()->plan->set_user_active_plan( $sub_id, $user_id, $plan_data['id'] );

		// NOTIFICATION
		mlmFire()->notif->send_user_sms( $user_id, 'plan_activated', array(
			'plan_id'	=> $plan_data['id'],
			'plan_name'	=> $plan_data['name']
		) );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'plan_id'	=> $plan_data['id']
			),
		) );
	}


	public function callback_subscribe_history( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$list		= array();
		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$per		= isset( $data['count'] ) ? absint( $data['count'] ) : 15;
		$start		= intval( ($paged - 1) * $per );
		$results	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE user_id = %d ORDER BY id DESC LIMIT %d, %d",
			array( $user_id, $start, $per ),
			'subscribe'
		);

		$all_pages	= mlmFire()->db->count_query_rows(
			"SELECT COUNT(id) FROM {TABLE} WHERE user_id = %d",
			array( $user_id ),
			'subscribe'
		);

		if( ! empty( $results ) )
		{
			foreach( $results as $subscribe )
			{
				$list[]	= array(
					'id'		=> $subscribe->id,
					'status'	=> $subscribe->status,
					'status_fa'	=> mlmFire()->plan->get_subscription_status( $subscribe->status ),
					'plan_data'	=> maybe_unserialize( $subscribe->plan_data ),
					'date'		=> date_i18n( 'j F Y', strtotime( $subscribe->date ) ),
					'time'		=> date_i18n( 'H:i', strtotime( $subscribe->date ) ),
				);
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'count'			=> count( $list ),
				'all_pages'		=> $all_pages,
				'items'			=> $list
			),
		) );
	}


	public function callback_subscribe_list( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$list		= array();
		$plan_id	= isset( $data['plan_id'] ) ? absint( $data['plan_id'] ) : '';
		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$per		= isset( $data['count'] ) ? absint( $data['count'] ) : 15;
		$query		= new WP_Query( array(
			'post_type' 		=> 'product',
			'post_status'		=> 'publish',
			'posts_per_page'	=> $per,
			'paged'				=> $paged,
			'tax_query'			=> array( array(
				'taxonomy'	=> 'plans',
				'field'		=> 'term_id',
				'terms'		=> $plan_id,
			) )
		) );

		if( $query->have_posts() )
		{
			while( $query->have_posts() )
			{
				$query->the_post();

				$post_id	= get_the_ID();
				$list[]		= array(
					'id'		=> $post_id,
					'title'		=> get_the_title(),
					'image'		=> mlm_image_url( $post_id, 'thumbnail', false ),
					'access'	=> mlmFire()->plan->check_user_access( $post_id, $user_id )
				);
			}

			wp_reset_postdata();
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'count'			=> count( $list ),
				'all_pages'		=> isset( $query->max_num_pages ) ? absint( $query->max_num_pages ) : 0,
				'items'			=> $list
			),
		) );
	}


	public function callback_check_access( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
		$user_obj	= get_userdata( $user_id );
		$user_email	= $user_obj->user_email;
		$access		= mlmFire()->plan->check_user_access( $post_id, $user_id );
		$purchased	= false;

		if( function_exists('wc_customer_bought_product') && wc_customer_bought_product( $user_email, $user_id, $post_id ) )
		{
			$purchased	= true;
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'purchased'		=> $purchased,
				'access'		=> $access
			),
		) );
	}


	public function callback_follow_user( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$vendor_id	= isset( $data['vendor_id'] ) ? absint( $data['vendor_id'] ) : 0;
		$followed	= mlmFire()->follow->do_user_follows( $user_id, $vendor_id );
		$mode		= ( $followed ) ? 'unfollow' : 'follow';

		if( ! mlm_user_exists( $vendor_id ) )
		{
			return $this->get_error('invalid_user_id');
		}

		if( $vendor_id == $user_id )
		{
			return $this->get_error('cant_follow_yourself');
		}

		if( $mode == 'unfollow' )
		{
			$result		= mlmFire()->follow->unfollow_user( $user_id, $vendor_id );
		}
		else
		{
			$result		= mlmFire()->follow->follow_user( $user_id, $vendor_id );
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'vendor_id'	=> $vendor_id
			),
		) );
	}


	public function callback_rate_post( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
		$rate		= isset( $data['rate'] ) ? absint( $data['rate'] ) : 0;

		if( ! mlm_post_exists( $post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$result	= mlmFire()->rating->add_rate( $post_id, $rate );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'average'	=> mlmFire()->rating->get_average( $post_id ),
				'total'		=> mlmFire()->rating->total_count( $post_id ),
			),
		) );
	}


	public function callback_post_articles( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;

		if( ! mlm_post_exists( $post_id ) || ! mlm_check_course( $post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$permission 	= false;
		$user_obj		= get_userdata( $user_id );
		$user_email		= $user_obj->user_email;
		$access			= mlmFire()->plan->check_user_access( $post_id, $user_id );
		$list			= array();
		$chapters_obj	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE post_id = %d AND parent_id = %d ORDER BY priority ASC",
			array( $post_id, 0 ),
			'course'
		);

		if( $access || ( function_exists('wc_customer_bought_product') && wc_customer_bought_product( $user_email, $user_id, $post_id ) ) )
		{
			$permission	= true;
		}

		if( ! empty( $chapters_obj ) )
		{
			foreach( $chapters_obj as $chapter )
			{
				$childs			= array();
				$chapter_data	= maybe_unserialize( $chapter->course_data );
				$image_atts		= wp_get_attachment_image_src( $chapter_data['image_id'], 'thumbnail' );
				$image_url		= ( $image_atts ) ? $image_atts[0] : IMAGES . '/no-thumbnail.png';
				$lesson_obj		= mlmFire()->db->query_rows(
					"SELECT * FROM {TABLE} WHERE parent_id = %d ORDER BY priority ASC",
					array( $chapter->id ),
					'course'
				);

				if( ! empty( $lesson_obj ) )
				{
					foreach( $lesson_obj as $lesson )
					{
						$lesson_data	= maybe_unserialize( $lesson->course_data );
						$linksss		= ( isset($lesson_data['links']) && is_array($lesson_data['links']) ) ? $lesson_data['links'] : array();
						$permis			= ( $permission || $lesson_data['status'] == 'free' ) ? true : false;
						$childs[]		= array(
							'id'		=> $lesson->id,
							'title'		=> $lesson_data['title'],
							'text'		=> $lesson_data['text'],
							'content'	=> $permis ? $lesson_data['content'] : '',
							'links'		=> $permis ? $linksss : array(),
							'access'	=> $permis,
						);
					}
				}

				$list[]	= array(
					'id'		=> $chapter->id,
					'title'		=> $chapter_data['title'],
					'text'		=> $chapter_data['text'],
					'image'		=> $image_url,
					'lessons'	=> $childs
				);
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'items'			=> $list,
				'permission'	=> $permission
			),
		) );
	}


	public function callback_post_author( $data )
	{
		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;

		if( ! mlm_post_exists( $post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$author_id		= get_post_field( 'post_author', $post_id );
		$author_obj		= get_userdata( $author_id );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'author_id'		=> $author_id,
				'author_name'	=> $author_obj->display_name,
				'author_avatar'	=> get_avatar_url( $author_id )
			),
		) );
	}


	public function callback_comments_query( $data )
	{
		/*$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}*/

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;

		if( ! mlm_post_exists( $post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$list		= array();
		$comments	= get_comments( array(
			'post_id'		=> $post_id,
			'post_type'		=> 'any',
			'status'		=> array( 'approve', 1 ),
			//'type'			=> 'comment',
			//'hierarchical'	=> true,
			'count'			=> false,
		) );

		if( ! is_array( $comments ) || ! count( $comments ) )
		{
			return $this->get_error('no_items_found');
		}

		foreach( $comments as $comment )
		{
			$post_id		= $comment->comment_post_ID;
			$commenter		= isset( $comment->user_id ) ? $comment->user_id : '';
			$label			= array();
			$label_class	= array();

			if( mlm_user_exists( $commenter ) )
			{
				$verified	= mlmFire()->dashboard->get_account_status( $commenter );

				if( $verified )
				{
					$label_class[]	= 'verified';
				}

				if( user_can( $commenter, 'manage_options' ) )
				{
					$label[]		= 'مدیر کل';
					$label_class[]	= 'admin';
				}
				elseif( $commenter == get_post_field( 'post_author', $post_id ) && 'product' == get_post_field( 'post_type', $post_id ) )
				{
					$label[]		= 'فروشنده';
					$label_class[]	= 'vendor';
				}
				elseif( $commenter == get_post_field( 'post_author', $post_id ) )
				{
					$label[]		= 'نویسنده';
					$label_class[]	= 'author';
				}
				elseif(
					'product' == get_post_field( 'post_type', $post_id ) &&
					function_exists('wc_customer_bought_product') &&
					wc_customer_bought_product( $comment->comment_author_email, $commenter, $post_id )
				)
				{
					$label[]		= 'خریدار محصول';
					$label_class[]	= 'customer';
				}
			}

			$list[]	= array(
				'id'			=> $comment->comment_ID,
				'name'			=> $comment->comment_author,
				'email'			=> $comment->comment_author_email,
				'date'			=> date_i18n( 'Y/m/d', strtotime( $comment->comment_date_gmt ) ),
				'time'			=> date_i18n( 'H:i', strtotime( $comment->comment_date_gmt ) ),
				'comment'		=> $comment->comment_content,
				'parent'		=> $comment->comment_parent,
				'label'			=> $label,
				'label_class'	=> $label_class,
				'avatar'		=> get_avatar_url( $commenter )
			);
		}

		return rest_ensure_response( array(
			'status'		=> 13,
			'data'			=> array(
				'count'			=> count( $list ),
				'items'			=> $list
			),
		) );
	}


	public function callback_new_comment( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		/*if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}*/

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
		$name		= isset( $data['name'] ) ? sanitize_text_field( $data['name'] ) : '';
		$email		= isset( $data['email'] ) ? sanitize_text_field( $data['email'] ) : '';
		$comment	= isset( $data['comment'] ) ? esc_textarea( $data['comment'] ) : '';
		$parent		= isset( $data['parent'] ) ? absint( $data['parent'] ) : 0;

		if( ! mlm_post_exists( $post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		if( empty( $name ) || empty( $comment ) || empty( $email ) )
		{
			return $this->get_error('required_fields');
		}

		$comment_id = wp_insert_comment( array(
			'comment_post_ID'		=> $post_id,
			'comment_content'		=> $comment,
			'user_id'				=> $user_id,
			'comment_author'		=> $name,
			'comment_author_email'	=> $email,
			'comment_parent'		=> $parent,
			'comment_approved'		=> 0
		) );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'comment_id'	=> $comment_id,
			),
		) );
	}


	public function callback_get_my_posts( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ): 1;
		$count		= isset( $data['count'] ) ? absint( $data['count'] ): 10;
		$search		= isset( $data['search'] ) ? sanitize_text_field( $data['search'] ): '';
		$list		= array();
		$query		= new WP_Query( array(
			'post_type' 		=> 'post',
			'author'			=> $user_id,
			'post_status'		=> array( 'publish', 'pending' ),
			'posts_per_page'	=> $count,
			'paged'				=> $paged,
			's'					=> $search
		) );

		if( $query->have_posts() )
		{
			while( $query->have_posts() )
			{
				$query->the_post();

				$post_id	= get_the_ID();

				$list[]	= array(
					'id'		=> $post_id,
					'status'	=> get_post_field( 'post_status', $post_id ),
					'image_url'	=> mlm_image_url( $post_id, 'thumbnail', false ),
					'title'		=> get_the_title( $post_id ),
					'views'		=> mlm_get_post_views( $post_id ),
				);
			}

			wp_reset_postdata();
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'			=> array(
				'count'			=> count( $list ),
				'all_pages'		=> isset( $query->max_num_pages ) ? absint( $query->max_num_pages ) : 0,
				'items'			=> $list
			),
		) );
	}


	public function callback_get_post( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;

		if( ! mlm_post_exists( $post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$author		= get_post_field( 'post_author', $post_id );
		$type		= get_post_field( 'post_type', $post_id );
		$status		= get_post_field( 'post_status', $post_id );

		if( $author != $user_id || $type != 'post' || ( $status != 'publish' && $status != 'pending' ) )
		{
			return $this->get_error('not_allowed');
		}

		$post = array(
			'title'		=> get_the_title( $post_id ),
			'content'	=> get_post_field( 'post_content', $post_id ),
			'cats'		=> mlm_category_list( $post_id, 'category' ),
			'tags'		=> mlm_category_list( $post_id, 'post_tag' ),
			'image_id'	=> get_post_meta( $post_id, '_thumbnail_id', true ),
			'image_url'	=> null,
			'reject'	=> get_post_meta( $post_id, 'mlm_reject', true ),
		);

		if( ! empty( $post['image_id'] ) && wp_get_attachment_url( $post['image_id'] ) )
		{
			$post['image_url']	= wp_get_attachment_url( $post['image_id'] );
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $post,
		) );
	}


	public function callback_new_post( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
        $title		= isset( $data['title'] ) ? sanitize_text_field( $data['title'] ) : '';
        $content	= isset( $data['content'] ) ? wp_filter_post_kses( $data['content'] ) : '';
        $thumb		= isset( $data['image_id'] ) ? absint( $data['image_id'] ) : 0;
        $cats		= isset( $data['cats'] ) ? mlm_sanitize_array( $data['cats'] ) : '';
        $tags		= isset( $data['tags'] ) ? mlm_sanitize_array( $data['tags'] ) : '';

		if( empty( $title ) || empty( $content ) || empty( $cats ) || empty( $tags ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! user_can( $user_id, 'upload_files' ) )
		{
			return $this->get_error('not_allowed');
		}

		if( $post_id )
		{
			if( ! mlm_post_exists( $post_id ) )
			{
				return $this->get_error('invalid_post_id');
			}

			$author		= get_post_field( 'post_author', $post_id );
			$type		= get_post_field( 'post_type', $post_id );
			$status		= get_post_field( 'post_status', $post_id );

			if( $author != $user_id || $type != 'post' || ( $status != 'publish' && $status != 'pending' ) )
			{
				return $this->get_error('not_allowed');
			}

			// UPDATE
			$update_post = array(
				'ID'			=> $post_id,
				'post_title'	=> wp_strip_all_tags( $title ),
				'post_content'	=> $content,
				'post_status'	=> 'pending',
			);
			wp_update_post( $update_post );
		}
		else
		{
			// SUBMIT
            $submit_post = array(
                'post_title'	=> wp_strip_all_tags( $title ),
                'post_content'	=> $content,
                'post_status'	=> 'pending',
                'post_type'		=> 'post',
                'post_author'	=> $user_id
            );

            $post_id	= wp_insert_post( $submit_post );
		}

		if( ! empty( $thumb ) && wp_get_attachment_url( $thumb ) )
        {
            update_post_meta( $post_id, '_thumbnail_id', $thumb );
        }
        else
        {
            delete_post_meta( $post_id, '_thumbnail_id' );
        }

        mlm_update_tax( $post_id, 'category', $cats );
        mlm_update_tax( $post_id, 'post_tag', $tags );

        // NOTIFICATION
        mlmFire()->notif->send_user_mail( $user_id, 'post_moderation', array(
            'post_id'	=> $post_id
        ) );
        mlmFire()->notif->send_admin_mail( 'post_moderation', array(
            'post_id'	=> $post_id
        ) );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'post_id'	=> $post_id,
			),
		) );
	}


	public function callback_get_my_products( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ): 1;
		$count		= isset( $data['count'] ) ? absint( $data['count'] ): 10;
		$status		= isset( $data['status'] ) ? sanitize_text_field( $data['status'] ): '';
		$search		= isset( $data['search'] ) ? sanitize_text_field( $data['search'] ): '';
		$list		= array();
		$args		= array(
			'post_type' 		=> 'product',
			'author'			=> $user_id,
			'posts_per_page'	=> $count,
			'paged'				=> $paged,
			's'					=> $search
		);

		switch( $status )
		{
			case 'publish_p':
				$args['post_status']	= 'publish';
				$args['meta_query']		= array(
					'relation'	=> 'OR',
					array(
						'key'		=> 'mlm_is_course',
						'value'		=> 'yes',
						'compare'	=> '!=',
					),
					array(
						'key'		=> 'mlm_is_course',
						'compare'	=> 'NOT EXISTS',
					),
				);
				break;

			case 'pending_p':
				$args['post_status']	= 'pending';
				$args['meta_query']		= array(
					'relation'	=> 'OR',
					array(
						'key'		=> 'mlm_is_course',
						'value'		=> 'yes',
						'compare'	=> '!=',
					),
					array(
						'key'		=> 'mlm_is_course',
						'compare'	=> 'NOT EXISTS',
					),
				);
				break;

			case 'draft_p':
				$args['post_status']	= 'draft';
				$args['meta_query']		= array(
					'relation'	=> 'OR',
					array(
						'key'		=> 'mlm_is_course',
						'value'		=> 'yes',
						'compare'	=> '!=',
					),
					array(
						'key'		=> 'mlm_is_course',
						'compare'	=> 'NOT EXISTS',
					),
				);
				break;

			case 'publish_c':
				$args['post_status']	= 'publish';
				$args['meta_query']		= array( array(
					'key'		=> 'mlm_is_course',
					'value'		=> 'yes',
					'compare'	=> '=',
				) );
				break;

			case 'pending_c':
				$args['post_status']	= 'pending';
				$args['meta_query']		= array( array(
					'key'		=> 'mlm_is_course',
					'value'		=> 'yes',
					'compare'	=> '=',
				) );
				break;

			case 'draft_c':
				$args['post_status']	= 'draft';
				$args['meta_query']		= array( array(
					'key'		=> 'mlm_is_course',
					'value'		=> 'yes',
					'compare'	=> '=',
				) );
				break;

			default:
				$args['post_status']	= array( 'publish', 'pending', 'draft' );
				break;
		}

		$query		= new WP_Query( $args );

		if( $query->have_posts() )
		{
			while( $query->have_posts() )
			{
				$query->the_post();

				$post_id	= get_the_ID();
				$is_file	= get_post_meta( $post_id, '_downloadable', true );
				$mlm_reject	= get_post_meta( $post_id, 'mlm_reject', true );

				$list[]	= array(
					'id'		=> $post_id,
					'status'	=> get_post_field( 'post_status', $post_id ),
					'image_url'	=> mlm_image_url( $post_id, 'thumbnail', false ),
					'title'		=> get_the_title( $post_id ),
					'price'		=> mlm_get_product_price( $post_id ),
					'views'		=> mlm_get_post_views( $post_id ),
					'sales'		=> get_post_meta( $post_id, 'total_sales', true ),
					'is_course' => ( mlm_check_course( $post_id ) ) ? true : false,
					'is_file'	=> ( $is_file != 'yes' ) ? false : true,
					'moderated'	=> ( $mlm_reject ) ? true : false,
				);
			}

			wp_reset_postdata();
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'			=> array(
				'count'			=> count( $list ),
				'all_pages'		=> isset( $query->max_num_pages ) ? absint( $query->max_num_pages ) : 0,
				'items'			=> $list
			),
		) );
	}


	public function callback_get_product( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id		= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;

		if( ! mlm_post_exists( $post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$author		= get_post_field( 'post_author', $post_id );
		$type		= get_post_field( 'post_type', $post_id );
		$status		= get_post_field( 'post_status', $post_id );

		if( $author != $user_id || $type != 'product' || ! in_array( $status, array( 'publish', 'pending', 'draft' ) ) )
		{
			return $this->get_error('not_allowed');
		}

		$types			= mlmFire()->wp_admin->supported_file_types();
		$fields_type	= mlm_custom_fields_type();
		$product		= array(
			'title'			=> get_the_title( $post_id ),
			'content'		=> get_post_field( 'post_content', $post_id ),
			'cats'			=> mlm_category_list( $post_id, 'product_cat' ),
			'tags'			=> mlm_category_list( $post_id, 'product_tag' ),
			'image_id'		=> get_post_meta( $post_id, '_thumbnail_id', true ),
			'image_url'		=> null,
			'reject'		=> get_post_meta( $post_id, 'mlm_reject', true ),
			'price'			=> get_post_meta( $post_id, '_regular_price', true ),
			'sale_price'	=> get_post_meta( $post_id, '_sale_price', true ),
			'files'			=> get_post_meta( $post_id, '_downloadable_files', true ),
			'ref_percent'	=> get_post_meta( $post_id, 'mlm_ref_value', true ),
			'button_text'	=> get_post_meta( $post_id, 'mlm_button_text', true ),
			'button_link'	=> get_post_meta( $post_id, 'mlm_button_link', true ),
			'button2_text'	=> get_post_meta( $post_id, 'mlm_button_2_text', true ),
			'button2_link'	=> get_post_meta( $post_id, 'mlm_button_2_link', true ),
			'image_thumb'	=> get_post_meta( $post_id, 'mlm_image_thumb', true ),
			'image_one'		=> get_post_meta( $post_id, 'mlm_image_one', true ),
			'image_two'		=> get_post_meta( $post_id, 'mlm_image_two', true ),
			'teacher_name'	=> get_post_meta( $post_id, 'mlm_teacher_name', true ),
			'teacher_image'	=> get_post_meta( $post_id, 'mlm_teacher_image', true ),
			'teacher_bio'	=> get_post_meta( $post_id, 'mlm_teacher_bio', true ),
			'course_video'	=> get_post_meta( $post_id, 'mlm_course_video', true ),
			'course_fill'	=> get_post_meta( $post_id, 'mlm_course_fill', true ),
			'weight'		=> get_post_meta( $post_id, '_weight', true ),
			'quantity'		=> get_post_meta( $post_id, '_stock', true ),
		);

		if( ! empty( $product['image_id'] ) && wp_get_attachment_url( $product['image_id'] ) )
		{
			$product['image_url']	= wp_get_attachment_url( $product['image_id'] );
		}

		if( $fields_type == 'custom' )
		{
			$product['fields']	= get_post_meta( $post_id, 'mlm_saved_fields', true );
		}
		else
		{
			$product['file_type']		= get_post_meta( $post_id, 'mlm_file_type', true );
			$product['page_count']		= get_post_meta( $post_id, 'mlm_page_count', true );
			$product['part_count']		= get_post_meta( $post_id, 'mlm_part_count', true );
			$product['file_author']		= get_post_meta( $post_id, 'mlm_file_author', true );
			$product['file_size']		= get_post_meta( $post_id, 'mlm_file_size', true );
			$product['file_format']		= get_post_meta( $post_id, 'mlm_file_format', true );
			$product['file_language']	= get_post_meta( $post_id, 'mlm_file_language', true );
			$product['file_step']		= get_post_meta( $post_id, 'mlm_file_step', true );
		}
		
		if( $stock == 'no' )
        {
            update_post_meta( $post_id, '_stock', 0 );
            update_post_meta( $post_id, '_stock_status', 'outofstock' );
        }
        else
        {
            update_post_meta( $post_id, '_stock_status', 'instock' );
        }

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $product,
		) );
	}


	public function callback_new_file( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id		= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
        $title			= isset( $data['title'] ) ? sanitize_text_field( $data['title'] ) : '';
        $content		= isset( $data['content'] ) ? wp_filter_post_kses( $data['content'] ) : '';
        $thumb			= isset( $data['image_id'] ) ? absint( $data['image_id'] ) : 0;
        $cats			= isset( $data['cats'] ) ? mlm_sanitize_array( $data['cats'] ) : '';
		$tags			= isset( $data['tags'] ) ? mlm_sanitize_array( $data['tags'] ) : '';
		$percent		= isset( $data['ref_percent'] ) ? absint( $data['ref_percent'] ) : 0;
		$price			= isset( $data['price'] ) ? absint( $data['price'] ) : 0;
        $sale_price		= isset( $data['sale_price'] ) ? absint( $data['sale_price'] ) : '';
        $button_text	= isset( $data['button_text'] ) ? sanitize_text_field( $data['button_text'] ) : '';
        $button_link	= isset( $data['button_link'] ) ? sanitize_text_field( $data['button_link'] ) : '';
		$button2_text	= isset( $data['button2_text'] ) ? sanitize_text_field( $data['button2_text'] ) : '';
        $button2_link	= isset( $data['button2_link'] ) ? sanitize_text_field( $data['button2_link'] ) : '';
        $files			= isset( $data['files'] ) ? mlm_sanitize_array( $data['files'] ) : '';
		$image_thumb	= isset( $data['image_thumb'] ) ? esc_url( $data['image_thumb'] ) : '';
		$image_one		= isset( $data['image_one'] ) ? esc_url( $data['image_one'] ) : '';
        $image_two		= isset( $data['image_two'] ) ? esc_url( $data['image_two'] ) : '';
		$stock		    = isset( $data['stock'] ) ? sanitize_text_field( $data['stock'] ) : '';
		$fields_type	= mlm_custom_fields_type();

		if( $fields_type == 'custom' )
        {
            $fields		= isset( $data['fields'] ) ? mlm_sanitize_array( $data['fields'] ) : '';
        }
        else
        {
            $file_type	= isset( $data['file_type'] ) ? sanitize_text_field( $data['file_type'] ) : '';
            $count		= isset( $data['page_count'] ) ? sanitize_text_field( $data['page_count'] ) : '';
            $part		= isset( $data['part_count'] ) ? sanitize_text_field( $data['part_count'] ) : '';
            $author		= isset( $data['file_author'] ) ? sanitize_text_field( $data['file_author'] ) : '';
            $size		= isset( $data['file_size'] ) ? sanitize_text_field( $data['file_size'] ) : '';
            $format		= isset( $data['file_format'] ) ? sanitize_text_field( $data['file_format'] ) : '';
            $language	= isset( $data['file_language'] ) ? sanitize_text_field( $data['file_language'] ) : '';
            $step		= isset( $data['file_step'] ) ? sanitize_text_field( $data['file_step'] ) : '';
        }

		if( empty( $title ) || empty( $content ) || empty( $cats ) || empty( $tags ) || ! is_numeric( $price ) || ! is_numeric( $percent ) )
		{
			return $this->get_error('required_fields');
		}

		if( $fields_type == 'custom' && mlmFire()->dashboard->check_required_fields( $fields ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! is_array( $files ) || ! count( $files ) )
        {
        	return $this->get_error('no_file_uploaded');
        }

		if( ! user_can( $user_id, 'level_3' ) )
		{
			return $this->get_error('not_allowed');
		}

		if( $post_id )
		{
			if( ! mlm_post_exists( $post_id ) )
			{
				return $this->get_error('invalid_post_id');
			}

			$vendor		= get_post_field( 'post_author', $post_id );
			$type		= get_post_field( 'post_type', $post_id );
			$status		= get_post_field( 'post_status', $post_id );

			if( $vendor != $user_id || $type != 'product' || ! in_array( $status, array( 'publish', 'pending', 'draft' ) ) || mlm_check_course( $post_id ) )
			{
				return $this->get_error('not_allowed');
			}

			// UPDATE
			$update_post = array(
				'ID'			=> $post_id,
				'post_title'	=> wp_strip_all_tags( $title ),
				'post_content'	=> $content,
				'post_status'	=> 'pending',
			);
			wp_update_post( $update_post );
		}
		else
		{
			// SUBMIT
            $submit_post = array(
                'post_title'	=> wp_strip_all_tags( $title ),
                'post_content'	=> $content,
                'post_status'	=> 'pending',
                'post_type'		=> 'product',
                'post_author'	=> $user_id
            );

            $post_id	= wp_insert_post( $submit_post );

			update_post_meta( $post_id, 'total_sales', 0 );
		}

		if( ! empty( $thumb ) && wp_get_attachment_url( $thumb ) )
        {
            update_post_meta( $post_id, '_thumbnail_id', $thumb );
        }
        else
        {
            delete_post_meta( $post_id, '_thumbnail_id' );
        }

        mlm_update_tax( $post_id, 'product_cat', $cats );
        mlm_update_tax( $post_id, 'product_tag', $tags );

		if( ! empty( $sale_price ) )
        {
            update_post_meta( $post_id, '_price', $sale_price );
            update_post_meta( $post_id, '_sale_price', $sale_price );
        }
        else
        {
            update_post_meta( $post_id, '_price', $price );
            delete_post_meta( $post_id, '_sale_price' );
        }

		$dl_files	= array();
        foreach( $files as $f )
        {
            $md5_num	= md5( $f['file'] );
            $dl_files[$md5_num] = array(
                'id'	=> $md5_num,
                'name'	=> empty( $f['name'] ) ? __( 'Download file', 'mlm' ) : $f['name'],
                'file'	=> $f['file']
            );
        }

		update_post_meta( $post_id, '_downloadable_files', $dl_files );
        update_post_meta( $post_id, '_regular_price', $price );
        update_post_meta( $post_id, '_virtual', 'yes' );
        update_post_meta( $post_id, '_downloadable', 'yes' );
        update_post_meta( $post_id, 'mlm_ref_value', $percent );
        update_post_meta( $post_id, 'mlm_button_text', $button_text );
        update_post_meta( $post_id, 'mlm_button_link', $button_link );
        update_post_meta( $post_id, 'mlm_button_2_text', $button2_text );
        update_post_meta( $post_id, 'mlm_button_2_link', $button2_link );
        update_post_meta( $post_id, 'mlm_image_thumb', $image_thumb );
        update_post_meta( $post_id, 'mlm_image_one', $image_one );
        update_post_meta( $post_id, 'mlm_image_two', $image_two );

        if( $fields_type == 'custom' )
        {
            update_post_meta( $post_id, 'mlm_saved_fields', $fields );
        }
        else
        {
            update_post_meta( $post_id, 'mlm_file_type', $file_type );
            update_post_meta( $post_id, 'mlm_page_count', $count );
            update_post_meta( $post_id, 'mlm_part_count', $part );
            update_post_meta( $post_id, 'mlm_file_author', $author );
            update_post_meta( $post_id, 'mlm_file_size', $size );
            update_post_meta( $post_id, 'mlm_file_format', $format );
            update_post_meta( $post_id, 'mlm_file_language', $language );
            update_post_meta( $post_id, 'mlm_file_step', $step );
        }
		
		if( $stock == 'no' )
        {
            update_post_meta( $post_id, '_stock', 0 );
            update_post_meta( $post_id, '_stock_status', 'outofstock' );
        }
        else
        {
            update_post_meta( $post_id, '_stock_status', 'instock' );
        }

		mlmFire()->wallet->post_ref_amount( $post_id );

        // NOTIFICATION
        mlmFire()->notif->send_user_mail( $user_id, 'product_moderation', array(
            'post_id'	=> $post_id
        ) );
        mlmFire()->notif->send_admin_mail( 'product_moderation', array(
            'post_id'	=> $post_id
        ) );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'post_id'	=> $post_id,
			),
		) );
	}


	public function callback_new_course( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id		= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
        $title			= isset( $data['title'] ) ? sanitize_text_field( $data['title'] ) : '';
        $content		= isset( $data['content'] ) ? wp_filter_post_kses( $data['content'] ) : '';
        $thumb			= isset( $data['image_id'] ) ? absint( $data['image_id'] ) : 0;
        $cats			= isset( $data['cats'] ) ? mlm_sanitize_array( $data['cats'] ) : '';
		$tags			= isset( $data['tags'] ) ? mlm_sanitize_array( $data['tags'] ) : '';
		$percent		= isset( $data['ref_percent'] ) ? absint( $data['ref_percent'] ) : 0;
		$price			= isset( $data['price'] ) ? absint( $data['price'] ) : 0;
        $sale_price		= isset( $data['sale_price'] ) ? absint( $data['sale_price'] ) : '';
        $button_text	= isset( $data['button_text'] ) ? sanitize_text_field( $data['button_text'] ) : '';
        $button_link	= isset( $data['button_link'] ) ? sanitize_text_field( $data['button_link'] ) : '';
		$button2_text	= isset( $data['button2_text'] ) ? sanitize_text_field( $data['button2_text'] ) : '';
        $button2_link	= isset( $data['button2_link'] ) ? sanitize_text_field( $data['button2_link'] ) : '';
		$image_thumb	= isset( $data['image_thumb'] ) ? esc_url( $data['image_thumb'] ) : '';
		$image_one		= isset( $data['image_one'] ) ? esc_url( $data['image_one'] ) : '';
        $image_two		= isset( $data['image_two'] ) ? esc_url( $data['image_two'] ) : '';
		$stock		    = isset( $data['stock'] ) ? sanitize_text_field( $data['stock'] ) : '';
        $teacher_image	= isset( $data['teacher_image'] ) ? esc_url( $data['teacher_image'] ) : '';
        $teacher_name	= isset( $data['teacher_name'] ) ? sanitize_text_field( $data['teacher_name'] ) : '';
        $course_fill	= isset( $data['course_fill'] ) ? absint( $data['course_fill'] ) : '';
        $teacher_bio	= isset( $data['teacher_bio'] ) ? esc_textarea( $data['teacher_bio'] ) : '';
        $course_video	= isset( $data['course_video'] ) ? esc_textarea( $data['course_video'] ) : '';
		$fields_type	= mlm_custom_fields_type();

		if( $fields_type == 'custom' )
        {
            $fields		= isset( $data['fields'] ) ? mlm_sanitize_array( $data['fields'] ) : '';
        }
        else
        {
            $file_type	= isset( $data['file_type'] ) ? sanitize_text_field( $data['file_type'] ) : '';
            $count		= isset( $data['page_count'] ) ? sanitize_text_field( $data['page_count'] ) : '';
            $part		= isset( $data['part_count'] ) ? sanitize_text_field( $data['part_count'] ) : '';
            $author		= isset( $data['file_author'] ) ? sanitize_text_field( $data['file_author'] ) : '';
            $size		= isset( $data['file_size'] ) ? sanitize_text_field( $data['file_size'] ) : '';
            $format		= isset( $data['file_format'] ) ? sanitize_text_field( $data['file_format'] ) : '';
            $language	= isset( $data['file_language'] ) ? sanitize_text_field( $data['file_language'] ) : '';
            $step		= isset( $data['file_step'] ) ? sanitize_text_field( $data['file_step'] ) : '';
        }

		if( empty( $title ) || empty( $content ) || empty( $cats ) || empty( $tags ) || ! is_numeric( $price ) || ! is_numeric( $percent ) || empty( $teacher_name ) )
		{
			return $this->get_error('required_fields');
		}

		if( $fields_type == 'custom' && mlmFire()->dashboard->check_required_fields( $fields ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! user_can( $user_id, 'level_3' ) )
		{
			return $this->get_error('not_allowed');
		}

		if( $post_id )
		{
			if( ! mlm_post_exists( $post_id ) )
			{
				return $this->get_error('invalid_post_id');
			}

			$vendor		= get_post_field( 'post_author', $post_id );
			$type		= get_post_field( 'post_type', $post_id );
			$status		= get_post_field( 'post_status', $post_id );

			if( $vendor != $user_id || $type != 'product' || ! in_array( $status, array( 'publish', 'pending', 'draft' ) ) || mlm_check_course( $post_id ) )
			{
				return $this->get_error('not_allowed');
			}

			// UPDATE
			$update_post = array(
				'ID'			=> $post_id,
				'post_title'	=> wp_strip_all_tags( $title ),
				'post_content'	=> $content,
				'post_status'	=> 'pending',
			);
			wp_update_post( $update_post );
		}
		else
		{
			// SUBMIT
            $submit_post = array(
                'post_title'	=> wp_strip_all_tags( $title ),
                'post_content'	=> $content,
                'post_status'	=> 'pending',
                'post_type'		=> 'product',
                'post_author'	=> $user_id
            );

            $post_id	= wp_insert_post( $submit_post );

			update_post_meta( $post_id, 'total_sales', 0 );
            update_post_meta( $post_id, 'mlm_is_course', 'yes' );
		}

		if( ! empty( $thumb ) && wp_get_attachment_url( $thumb ) )
        {
            update_post_meta( $post_id, '_thumbnail_id', $thumb );
        }
        else
        {
            delete_post_meta( $post_id, '_thumbnail_id' );
        }

        mlm_update_tax( $post_id, 'product_cat', $cats );
        mlm_update_tax( $post_id, 'product_tag', $tags );

		if( ! empty( $sale_price ) )
        {
            update_post_meta( $post_id, '_price', $sale_price );
            update_post_meta( $post_id, '_sale_price', $sale_price );
        }
        else
        {
            update_post_meta( $post_id, '_price', $price );
            delete_post_meta( $post_id, '_sale_price' );
        }

        update_post_meta( $post_id, '_regular_price', $price );
        update_post_meta( $post_id, '_virtual', 'yes' );
        update_post_meta( $post_id, 'mlm_ref_value', $percent );
        update_post_meta( $post_id, 'mlm_button_text', $button_text );
        update_post_meta( $post_id, 'mlm_button_link', $button_link );
        update_post_meta( $post_id, 'mlm_button_2_text', $button2_text );
        update_post_meta( $post_id, 'mlm_button_2_link', $button2_link );
        update_post_meta( $post_id, 'mlm_image_thumb', $image_thumb );
        update_post_meta( $post_id, 'mlm_image_one', $image_one );
        update_post_meta( $post_id, 'mlm_image_two', $image_two );
		update_post_meta( $post_id, 'mlm_teacher_image', $teacher_image );
        update_post_meta( $post_id, 'mlm_teacher_name', $teacher_name );
        update_post_meta( $post_id, 'mlm_course_fill', $course_fill );
        update_post_meta( $post_id, 'mlm_teacher_bio', $teacher_bio );
        update_post_meta( $post_id, 'mlm_course_video', $course_video );

        if( $fields_type == 'custom' )
        {
            update_post_meta( $post_id, 'mlm_saved_fields', $fields );
        }
        else
        {
            update_post_meta( $post_id, 'mlm_file_type', $file_type );
            update_post_meta( $post_id, 'mlm_page_count', $count );
            update_post_meta( $post_id, 'mlm_part_count', $part );
            update_post_meta( $post_id, 'mlm_file_author', $author );
            update_post_meta( $post_id, 'mlm_file_size', $size );
            update_post_meta( $post_id, 'mlm_file_format', $format );
            update_post_meta( $post_id, 'mlm_file_language', $language );
            update_post_meta( $post_id, 'mlm_file_step', $step );
        }
		
		if( $stock == 'no' )
        {
            update_post_meta( $post_id, '_stock', 0 );
            update_post_meta( $post_id, '_stock_status', 'outofstock' );
        }
        else
        {
            update_post_meta( $post_id, '_stock_status', 'instock' );
        }

		mlmFire()->wallet->post_ref_amount( $post_id );

        // NOTIFICATION
        mlmFire()->notif->send_user_mail( $user_id, 'product_moderation', array(
            'post_id'	=> $post_id
        ) );
        mlmFire()->notif->send_admin_mail( 'product_moderation', array(
            'post_id'	=> $post_id
        ) );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'post_id'	=> $post_id,
			),
		) );
	}


	public function callback_new_physical( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id		= isset( $data['post_id'] ) ? absint( $data['post_id'] ) : 0;
        $title			= isset( $data['title'] ) ? sanitize_text_field( $data['title'] ) : '';
        $content		= isset( $data['content'] ) ? wp_filter_post_kses( $data['content'] ) : '';
        $thumb			= isset( $data['image_id'] ) ? absint( $data['image_id'] ) : 0;
        $cats			= isset( $data['cats'] ) ? mlm_sanitize_array( $data['cats'] ) : '';
		$tags			= isset( $data['tags'] ) ? mlm_sanitize_array( $data['tags'] ) : '';
		$percent		= isset( $data['ref_percent'] ) ? absint( $data['ref_percent'] ) : 0;
		$price			= isset( $data['price'] ) ? absint( $data['price'] ) : 0;
        $sale_price		= isset( $data['sale_price'] ) ? absint( $data['sale_price'] ) : '';
        $button_text	= isset( $data['button_text'] ) ? sanitize_text_field( $data['button_text'] ) : '';
        $button_link	= isset( $data['button_link'] ) ? sanitize_text_field( $data['button_link'] ) : '';
		$button2_text	= isset( $data['button2_text'] ) ? sanitize_text_field( $data['button2_text'] ) : '';
        $button2_link	= isset( $data['button2_link'] ) ? sanitize_text_field( $data['button2_link'] ) : '';
		$image_thumb	= isset( $data['image_thumb'] ) ? esc_url( $data['image_thumb'] ) : '';
		$image_one		= isset( $data['image_one'] ) ? esc_url( $data['image_one'] ) : '';
        $image_two		= isset( $data['image_two'] ) ? esc_url( $data['image_two'] ) : '';
		$stock		    = isset( $data['stock'] ) ? sanitize_text_field( $data['stock'] ) : '';
        $weight			= isset( $data['weight'] ) ? absint( $data['weight'] ) : '';
        $quantity		= isset( $data['quantity'] ) ? absint( $data['quantity'] ) : '';
		$fields_type	= mlm_custom_fields_type();

		if( $fields_type == 'custom' )
        {
            $fields		= isset( $data['fields'] ) ? mlm_sanitize_array( $data['fields'] ) : '';
        }
        else
        {
            $file_type	= isset( $data['file_type'] ) ? sanitize_text_field( $data['file_type'] ) : '';
            $count		= isset( $data['page_count'] ) ? sanitize_text_field( $data['page_count'] ) : '';
            $part		= isset( $data['part_count'] ) ? sanitize_text_field( $data['part_count'] ) : '';
            $author		= isset( $data['file_author'] ) ? sanitize_text_field( $data['file_author'] ) : '';
            $size		= isset( $data['file_size'] ) ? sanitize_text_field( $data['file_size'] ) : '';
            $format		= isset( $data['file_format'] ) ? sanitize_text_field( $data['file_format'] ) : '';
            $language	= isset( $data['file_language'] ) ? sanitize_text_field( $data['file_language'] ) : '';
            $step		= isset( $data['file_step'] ) ? sanitize_text_field( $data['file_step'] ) : '';
        }

		if( empty( $title ) || empty( $content ) || empty( $cats ) || empty( $tags ) || ! is_numeric( $price ) || ! is_numeric( $percent ) || empty( $weight ) || empty( $quantity ) )
		{
			return $this->get_error('required_fields');
		}

		if( $fields_type == 'custom' && mlmFire()->dashboard->check_required_fields( $fields ) )
		{
			return $this->get_error('required_fields');
		}

		if( ! user_can( $user_id, 'level_3' ) )
		{
			return $this->get_error('not_allowed');
		}

		if( $post_id )
		{
			if( ! mlm_post_exists( $post_id ) )
			{
				return $this->get_error('invalid_post_id');
			}

			$vendor		= get_post_field( 'post_author', $post_id );
			$type		= get_post_field( 'post_type', $post_id );
			$status		= get_post_field( 'post_status', $post_id );

			if( $vendor != $user_id || $type != 'product' || ! in_array( $status, array( 'publish', 'pending', 'draft' ) ) || mlm_check_course( $post_id ) )
			{
				return $this->get_error('not_allowed');
			}

			// UPDATE
			$update_post = array(
				'ID'			=> $post_id,
				'post_title'	=> wp_strip_all_tags( $title ),
				'post_content'	=> $content,
				'post_status'	=> 'pending',
			);
			wp_update_post( $update_post );
		}
		else
		{
			// SUBMIT
            $submit_post = array(
                'post_title'	=> wp_strip_all_tags( $title ),
                'post_content'	=> $content,
                'post_status'	=> 'pending',
                'post_type'		=> 'product',
                'post_author'	=> $user_id
            );

            $post_id	= wp_insert_post( $submit_post );

			update_post_meta( $post_id, 'total_sales', 0 );
		}

		if( ! empty( $thumb ) && wp_get_attachment_url( $thumb ) )
        {
            update_post_meta( $post_id, '_thumbnail_id', $thumb );
        }
        else
        {
            delete_post_meta( $post_id, '_thumbnail_id' );
        }

        mlm_update_tax( $post_id, 'product_cat', $cats );
        mlm_update_tax( $post_id, 'product_tag', $tags );

		if( ! empty( $sale_price ) )
        {
            update_post_meta( $post_id, '_price', $sale_price );
            update_post_meta( $post_id, '_sale_price', $sale_price );
        }
        else
        {
            update_post_meta( $post_id, '_price', $price );
            delete_post_meta( $post_id, '_sale_price' );
        }

        update_post_meta( $post_id, '_regular_price', $price );
		update_post_meta( $post_id, '_virtual', 'no' );
        update_post_meta( $post_id, '_downloadable', 'no' );
        update_post_meta( $post_id, '_manage_stock', 'yes' );
        update_post_meta( $post_id, '_weight', $weight );
        update_post_meta( $post_id, '_stock', $quantity );
        update_post_meta( $post_id, '_stock_status', 'instock' );
        update_post_meta( $post_id, 'mlm_ref_value', $percent );
        update_post_meta( $post_id, 'mlm_button_text', $button_text );
        update_post_meta( $post_id, 'mlm_button_link', $button_link );
        update_post_meta( $post_id, 'mlm_button_2_text', $button2_text );
        update_post_meta( $post_id, 'mlm_button_2_link', $button2_link );
        update_post_meta( $post_id, 'mlm_image_thumb', $image_thumb );
        update_post_meta( $post_id, 'mlm_image_one', $image_one );
        update_post_meta( $post_id, 'mlm_image_two', $image_two );

        if( $fields_type == 'custom' )
        {
            update_post_meta( $post_id, 'mlm_saved_fields', $fields );
        }
        else
        {
            update_post_meta( $post_id, 'mlm_file_type', $file_type );
            update_post_meta( $post_id, 'mlm_page_count', $count );
            update_post_meta( $post_id, 'mlm_part_count', $part );
            update_post_meta( $post_id, 'mlm_file_author', $author );
            update_post_meta( $post_id, 'mlm_file_size', $size );
            update_post_meta( $post_id, 'mlm_file_format', $format );
            update_post_meta( $post_id, 'mlm_file_language', $language );
            update_post_meta( $post_id, 'mlm_file_step', $step );
        }
		
		if( $stock == 'no' )
        {
            update_post_meta( $post_id, '_stock', 0 );
            update_post_meta( $post_id, '_stock_status', 'outofstock' );
        }
        else
        {
            update_post_meta( $post_id, '_stock_status', 'instock' );
        }

		mlmFire()->wallet->post_ref_amount( $post_id );

        // NOTIFICATION
        mlmFire()->notif->send_user_mail( $user_id, 'product_moderation', array(
            'post_id'	=> $post_id
        ) );
        mlmFire()->notif->send_admin_mail( 'product_moderation', array(
            'post_id'	=> $post_id
        ) );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'post_id'	=> $post_id,
			),
		) );
	}


	public function callback_course_articles( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['course_id'] ) ? absint( $data['course_id'] ) : 0;

		if( ! mlm_post_exists( $post_id ) || ! mlm_check_course( $post_id ) )
		{
			return $this->get_error('not_allowed');
		}

		$author		= get_post_field( 'post_author', $post_id );
		$type		= get_post_field( 'post_type', $post_id );
		$status		= get_post_field( 'post_status', $post_id );

		if( $author != $user_id || $type != 'product' || ! in_array( $status, array( 'publish', 'pending', 'draft' ) ) )
		{
			return $this->get_error('not_allowed');
		}

		$list	= array();
		$query	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE post_id = %d AND parent_id = %d ORDER BY priority ASC",
			array( $post_id, 0 ),
			'course'
		);

		if( ! empty( $query ) )
		{
			foreach( $query as $chapter )
			{
				$chapter_data	= maybe_unserialize( $chapter->course_data );
				$image_atts		= wp_get_attachment_image_src( $chapter_data['image_id'], 'thumbnail' );
				$lessons		= array();
				$course_obj		= mlmFire()->db->query_rows(
					"SELECT * FROM {TABLE} WHERE parent_id = %d ORDER BY priority ASC",
					array( $chapter->id ),
					'course'
				);

				if( ! empty( $course_obj ) )
				{
					foreach( $course_obj as $lesson )
					{
						$lesson_data	= maybe_unserialize( $lesson->course_data );
						$links			= ( isset($lesson_data['links']) && is_array($lesson_data['links']) ) ? $lesson_data['links'] : array();
						$lessons[]		= array(
							'id'			=> $lesson->id,
							'title'			=> $lesson_data['title'],
							'text'			=> $lesson_data['text'],
							'content'		=> $lesson_data['content'],
							
							'link'			=> isset( $lesson_data['link'] ) ? $lesson_data['link'] : '',
							'description'	=> isset( $lesson_data['description'] ) ? $lesson_data['description'] : '',
							
							'links'			=> $links,
							'status'		=> $lesson_data['status'],
							'priority'		=> $lesson->priority,
						);
					}
				}

				$list[]		= array(
					'id'		=> $chapter->id,
					'title'		=> $chapter_data['title'],
					'text'		=> $chapter_data['text'],
					'image_id'	=> $chapter_data['image_id'],
					'image_url'	=> ( ! $image_atts ) ? '' : $image_atts[0],
					'priority'	=> $chapter->priority,
					'lessons'	=> $lessons,
				);
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'			=> array(
				'count'			=> count( $list ),
				'items'			=> $list
			),
		) );
	}


	public function callback_course_chapter_save( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['course_id'] ) ? absint( $data['course_id'] ) : 0;
		$chapter_id	= isset( $data['chapter_id'] ) ? absint( $data['chapter_id'] ) : '';
		$image_id	= isset( $data['image_id'] ) ? absint( $data['image_id'] ) : '';
		$number		= isset( $data['number'] ) ? absint( $data['number'] ) : '';
		$title		= isset( $data['title'] ) ? sanitize_text_field( $data['title'] ) : '';
		$desc		= isset( $data['desc'] ) ? sanitize_text_field( $data['desc'] ) : '';
        $author		= get_post_field( 'post_author', $post_id );
        $type		= get_post_field( 'post_type', $post_id );

		if( ! mlm_post_exists( $post_id ) || $type != 'product' )
		{
			return $this->get_error('not_allowed');
		}

		if( ! user_can( $user_id, 'moderate_comments') && $author != $user_id )
		{
			return $this->get_error('not_allowed');
		}

		if( empty( $number ) || empty( $title ) || empty( $desc ) )
		{
			return $this->get_error('required_fields');
		}

		if( $chapter_id )
		{
			$course_data	= array(
				'priority'		=> $number,
				'course_data'	=> array(
					'title'			=> $title,
					'text'			=> $desc,
					'image_id'		=> $image_id,
				),
			);

			mlmFire()->db->course_update( $chapter_id, $course_data );
		}
		else
		{
			$course_data	= array(
				'title'			=> $title,
				'text'			=> $desc,
				'image_id'		=> $image_id,
			);
			$chapter_id		= mlmFire()->db->course_record( $post_id, 0, $number, $course_data );
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'chapter_id'	=> $chapter_id
			),
		) );
	}


	public function callback_course_lesson_save( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= isset( $data['course_id'] ) ? absint( $data['course_id'] ) : 0;
		$lesson_id	= isset( $data['lesson_id'] ) ? absint( $data['lesson_id'] ) : '';
		$chapter_id	= isset( $data['chapter_id'] ) ? absint( $data['chapter_id'] ) : '';
		$number		= isset( $data['number'] ) ? absint( $data['number'] ) : '';
		$title		= isset( $data['title'] ) ? sanitize_text_field( $data['title'] ) : '';
		$desc		= isset( $data['desc'] ) ? sanitize_text_field( $data['desc'] ) : '';
		$status		= isset( $data['status'] ) ? sanitize_text_field( $data['status'] ) : '';
		$links		= isset( $data['links'] ) ? mlm_sanitize_array( $data['links'] ) : '';
        $author		= get_post_field( 'post_author', $post_id );
        $type		= get_post_field( 'post_type', $post_id );
		$course_obj	= mlmFire()->db->query_rows(
            "SELECT * FROM {TABLE} WHERE id = %d LIMIT %d",
            array( $chapter_id, 1 ),
            'course',
            true
        );

		if( ! mlm_post_exists( $post_id ) || $type != 'product' )
		{
			return $this->get_error('not_allowed');
		}

		if( ! user_can( $user_id, 'moderate_comments') && $author != $user_id )
		{
			return $this->get_error('not_allowed');
		}

		if( ! isset( $course_obj->post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		if( empty( $number ) || empty( $title ) || empty( $desc ) || empty( $status )  )
		{
			return $this->get_error('required_fields');
		}

		if( $lesson_id )
		{
			$course_data	= array(
				'priority'		=> $number,
				'course_data'	=> array(
					'title'			=> $title,
					'text'			=> $desc,
					'status'		=> $status,
					'links'			=> $links,
				),
			);

			mlmFire()->db->course_update( $lesson_id, $course_data );
		}
		else
		{
			$course_data	= array(
				'title'			=> $title,
				'text'			=> $desc,
				'status'		=> $status,
				'content'		=> '',
			);
			$lesson_id	= mlmFire()->db->course_record( $post_id, $chapter_id, $number, $course_data );
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'lesson_id'	=> $lesson_id
			),
		) );
	}


	public function callback_course_chapter_delete( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$chapter_id	= isset( $_POST['chapter_id'] ) ? absint( $_POST['chapter_id'] ) : '';
        $course_obj	= mlmFire()->db->query_rows(
            "SELECT * FROM {TABLE} WHERE id = %d LIMIT %d",
            array( $chapter_id, 1 ),
            'course',
            true
        );

		if( ! isset( $course_obj->post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$author		= get_post_field( 'post_author', $course_obj->post_id );

		if( ! user_can( $user_id, 'moderate_comments') && $author != $user_id )
		{
			return $this->get_error('not_allowed');
		}

		mlmFire()->db->course_delete( $chapter_id );

		$childs	= mlmFire()->db->query_rows(
			"SELECT id FROM {TABLE} WHERE parent_id = %d",
			array( $chapter_id ),
			'course'
		);

		if( ! empty( $childs ) )
		{
			foreach( $childs as $child )
			{
				mlmFire()->db->course_delete( $child->id );
			}
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'chapter_id'	=> $chapter_id
			),
		) );
	}


	public function callback_course_lesson_delete( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$lesson_id	= isset( $_POST['lesson_id'] ) ? absint( $_POST['lesson_id'] ) : '';
		$course_obj	= mlmFire()->db->query_rows(
            "SELECT * FROM {TABLE} WHERE id = %d LIMIT %d",
            array( $lesson_id, 1 ),
            'course',
            true
        );

		if( ! isset( $course_obj->post_id ) )
		{
			return $this->get_error('invalid_post_id');
		}

		$author		= get_post_field( 'post_author', $course_obj->post_id );

		if( ! user_can( $user_id, 'moderate_comments') && $author != $user_id )
		{
			return $this->get_error('not_allowed');
		}

		mlmFire()->db->course_delete( $lesson_id );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'lesson_id'	=> $lesson_id
			),
		) );
	}


	public function callback_upgrade_data( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id		= mlmFire()->dashboard->get_request_post_id( $user_id );
		$upgrade		= array(
			'status'	=> get_post_meta( $post_id, 'mlm_status', true ),
			'reject'	=> get_post_meta( $post_id, 'mlm_reject', true ),
			'gender'	=> get_post_meta( $post_id, 'mlm_gender', true ),
			'fname'		=> get_post_meta( $post_id, 'mlm_fname', true ),
			'lname'		=> get_post_meta( $post_id, 'mlm_lname', true ),
			'birth'		=> get_post_meta( $post_id, 'mlm_birth', true ),
			'melli'		=> get_post_meta( $post_id, 'mlm_melli', true ),
			'phone'		=> get_post_meta( $post_id, 'mlm_phone', true ),
			'address'	=> get_post_meta( $post_id, 'mlm_address', true ),
			'postal'	=> get_post_meta( $post_id, 'mlm_postal', true ),
			'melli_file'=> get_post_meta( $post_id, 'mlm_melli_file', true ),
			'shena_file'=> get_post_meta( $post_id, 'mlm_shena_file', true ),
		);

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $upgrade,
		) );
	}


	public function callback_step_one( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= mlmFire()->dashboard->get_request_post_id( $user_id );
		$status		= get_post_meta( $post_id, 'mlm_status', true );
		$gender		= isset( $data['gender'] ) ? sanitize_text_field( $data['gender'] ) : 'm';
        $fname		= isset( $data['fname'] ) ? sanitize_text_field( $data['fname'] ) : '';
        $lname		= isset( $data['lname'] ) ? sanitize_text_field( $data['lname'] ) : '';
        $birth		= isset( $data['birth'] ) ? sanitize_text_field( $data['birth'] ) : '';
        $melli		= isset( $data['melli'] ) ? sanitize_text_field( $data['melli'] ) : '';
        $address	= isset( $data['address'] ) ? esc_textarea( $data['address'] ) : '';
        $phone		= isset( $data['phone'] ) ? sanitize_text_field( $data['phone'] ) : '';
        $postal		= isset( $data['postal'] ) ? sanitize_text_field( $data['postal'] ) : '';
        $role		= isset( $data['role'] ) ? absint( $data['role'] ) : '';

		if( user_can( $user_id, 'moderate_comments' ) )
		{
			return $this->get_error('already_upgraded');
		}

		if( $status == 'ok' )
        {
			return $this->get_error('already_verified');
        }

		if( $status == 'wait' )
        {
			return $this->get_error('wait_for_result');
        }

		if( empty( $gender ) || empty( $fname ) || empty( $lname ) || empty( $birth ) || empty( $melli ) || empty( $address ) || empty( $phone ) || empty( $postal ) )
        {
            return $this->get_error('required_fields');
        }

		if( ! mlm_post_exists( $post_id ) )
		{
			$submit_post = array(
				'post_title'	=> wp_strip_all_tags( mlm_get_user_name( $user_id ) ),
				'post_status'	=> 'pending',
				'post_type'		=> 'mlm-requests',
				'post_author'	=> $user_id
			);

			$post_id	= wp_insert_post( $submit_post );

			update_post_meta( $post_id, 'mlm_role', $role );
		}
		else
		{
			$update_post = array(
				'ID'			=> $post_id,
				'post_status'	=> 'pending',
			);
			wp_update_post( $update_post );
		}

		update_post_meta( $post_id, 'mlm_gender', $gender );
		update_post_meta( $post_id, 'mlm_fname', $fname );
		update_post_meta( $post_id, 'mlm_lname', $lname );
		update_post_meta( $post_id, 'mlm_birth', $birth );
		update_post_meta( $post_id, 'mlm_melli', $melli );
		update_post_meta( $post_id, 'mlm_phone', $phone );
		update_post_meta( $post_id, 'mlm_address', $address );
		update_post_meta( $post_id, 'mlm_postal', $postal );
		update_post_meta( $post_id, 'mlm_status', 'nok' );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'submitted'	=> true
			),
		) );
	}


	public function callback_step_two( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id	= mlmFire()->dashboard->get_request_post_id( $user_id );
		$status		= get_post_meta( $post_id, 'mlm_status', true );
		$melli		= isset( $data['melli_file'] ) ? esc_url( $data['melli_file'] ) : '';
        $shena		= isset( $data['shena_file'] ) ? esc_url( $data['shena_file'] ) : '';

		if( user_can( $user_id, 'moderate_comments' ) )
		{
			return $this->get_error('already_upgraded');
		}

		if( $status == 'ok' )
        {
			return $this->get_error('already_verified');
        }

		if( $status == 'wait' )
        {
			return $this->get_error('wait_for_result');
        }

		if( empty( $melli ) || empty( $shena ) )
        {
            return $this->get_error('required_fields');
        }

		if( ! mlm_post_exists( $post_id ) )
		{
			$submit_post = array(
				'post_title'	=> wp_strip_all_tags( mlm_get_user_name( $user_id ) ),
				'post_status'	=> 'pending',
				'post_type'		=> 'mlm-requests',
				'post_author'	=> $user_id
			);

			$post_id	= wp_insert_post( $submit_post );

			update_post_meta( $post_id, 'mlm_role', 1 );
		}
		else
		{
			$update_post = array(
				'ID'			=> $post_id,
				'post_status'	=> 'pending',
			);
			wp_update_post( $update_post );
		}

		update_post_meta( $post_id, 'mlm_melli_file', $melli );
		update_post_meta( $post_id, 'mlm_shena_file', $shena );
		update_post_meta( $post_id, 'mlm_status', 'wait' );

		// NOTIFICATION
		mlmFire()->notif->send_admin_mail( 'upgrade', array(
			'user_id'	=> $user_id
		) );

		mlmFire()->notif->send_user_mail( $user_id, 'upgrade_request' );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> array(
				'submitted'	=> true
			),
		) );
	}


	public function callback_announces_list( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$paged			= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$count			= isset( $data['count'] ) ? absint( $data['count'] ) : 10;
		$status			= isset( $data['status'] ) ? sanitize_text_field( $data['status'] ) : '';
		$meta_query		= mlmFire()->announce->announce_meta_query( $user_id );
		$user_announces	= mlmFire()->announce->get_user_announces( $user_id );
		$args			= array(
			'post_type' 		=> 'mlm-announce',
			'post_status' 		=> 'publish',
			'posts_per_page'	=> $count,
			'paged'				=> $paged,
		);

		if( $meta_query )
		{
			$args['meta_query'] = $meta_query;
		}

		switch( $status )
		{
			case 'seen':

				if( $user_announces )
				{
					$args['post__in'] = $user_announces;
				}
				else
				{
					$args['post__in'] = array( 0 );
				}

				break;

			case 'unseen':

				if( $user_announces )
				{
					$args['post__not_in'] = $user_announces;
				}
				else
				{
					$args['post__not_in'] = array( 0 );
				}

				break;
		}

		$list		= array();
		$query		= new WP_Query( $args );

		if( $query->have_posts() )
		{
			while( $query->have_posts() )
			{
				$query->the_post();

				$list[] = array(
					'id'		=> get_the_ID(),
					'title'		=> get_the_title(),
					'status'	=> ( ! $user_announces || ( is_array( $user_announces ) && ! in_array( get_the_ID(), $user_announces ) ) ) ? 'unseen' : 'seen',
					'local_date'	=> date_i18n( 'Y/m/d H:i:s', get_the_time( 'U' ) ),
				);
			}

			wp_reset_postdata();
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'			=> array(
				'count'			=> count( $list ),
				'all_pages'		=> isset( $query->max_num_pages ) ? absint( $query->max_num_pages ) : 0,
				'items'			=> $list
			),
		) );
	}


	public function callback_announces_details( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$post_id		= isset( $data['announce_id'] ) ? absint( $data['announce_id'] ) : 0;
		$user_access	= mlmFire()->announce->check_user_access( $user_id, $post_id );

		if( ! $user_access )
		{
			return $this->get_error('not_allowed');
		}

		$announce	= array(
			'id'		=> $post_id,
			'title'		=> get_the_title( $post_id ),
			'content'	=> get_post_field( 'post_content', $post_id ),
		);

		mlmFire()->announce->announce_seen_by_user( $user_id, $post_id );

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'		=> $announce,
		) );
	}


	public function callback_links_list( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$paged			= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$count			= isset( $data['count'] ) ? absint( $data['count'] ) : 10;
		$mode			= isset( $data['mode'] ) ? sanitize_text_field( $data['mode'] ) : 'new';
		$vendor			= isset( $data['vendor'] ) ? absint( $data['vendor'] ) : 0;
		$search			= isset( $data['search'] ) ? sanitize_text_field( $data['search'] ) : '';

		if( ! mlm_user_exists( $vendor ) )
		{
			$vendor = 0;
		}

		if( $mode == 'per' )
		{
			$meta_key	= 'mlm_ref_amount';
			$orderby	= 'meta_value_num';
		}
		elseif( $mode == 'sell' )
		{
			$meta_key	= 'total_sales';
			$orderby	= 'meta_value_num';
		}
		else
		{
			$meta_key	= 0;
			$orderby	= 'date';
		}

		$list	= array();
		$query	= new WP_Query( array(
			'post_type' 	=> 'product',
			'post_status'	=> 'publish',
			'author'		=> $vendor,
			'meta_key'		=> $meta_key,
			'orderby'		=> $orderby,
			'order'			=> 'DESC',
			'paged'			=> $paged,
			'posts_per_page'=> $count,
			's'				=> $search,
			'meta_query'	=> array(
				array(
					'key'		=> 'mlm_ref_value',
					'value'		=> 0,
					'type'		=> 'numeric',
					'compare'	=> '>',
				)
			),
		) );

		if( $query->have_posts() )
		{
			while( $query->have_posts() )
			{
				$query->the_post();

				$post_id	= get_the_ID();
				$list[]		= array(
					'title'			=> get_the_title( $post_id ),
					'image_url'		=> mlm_image_url( $post_id, 'thumbnail', false ),
					'price'			=> get_post_meta( $post_id, '_regular_price', true ),
					'sale_price'	=> get_post_meta( $post_id, '_sale_price', true ),
					'ref_link'		=> mlmFire()->referral->add_ref_to_url( wp_get_shortlink() ),
					'ref_amount'	=> mlmFire()->wallet->post_ref_amount( $post_id ),
					'ref_percent'	=> get_post_meta( $post_id, 'mlm_ref_value', true ),
				);
			}

			wp_reset_postdata();
		}

		return rest_ensure_response( array(
			'status'	=> 13,
			'data'			=> array(
				'count'			=> count( $list ),
				'all_pages'		=> isset( $query->max_num_pages ) ? absint( $query->max_num_pages ) : 0,
				'items'			=> $list
			),
		) );
	}


	public function callback_referrals_list( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$list		= array();
		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$count		= isset( $data['count'] ) ? absint( $data['count'] ) : 10;
		$start		= intval( ($paged - 1) * $count );
		$results	= mlmFire()->db->query_rows(
			"SELECT * FROM {TABLE} WHERE ref_user_id = %d ORDER BY id DESC LIMIT %d, %d",
			array( $user_id, $start, $count ),
			'referral'
		);
		$all_pages	= mlmFire()->db->count_query_rows(
			"SELECT COUNT(id) FROM {TABLE} WHERE ref_user_id = %d",
			array( $user_id ),
			'referral'
		);

		mlmFire()->referral->add_ref_to_url();

		if( ! empty( $results ) )
		{
			foreach( $results as $item )
			{
				$list[]		= array(
					'id'			=> $item->id,
					'source'		=> $item->user_host,
					'landing'		=> $item->user_url,
					'user_ip'		=> $item->user_ip,
					'valid'			=> ( $item->invalid ) ? false : true,
					'date'			=> $item->date,
					'local_date'	=> date_i18n( 'Y/m/d H:i:s', strtotime( $item->date ) ),
				);
			}
		}

		return rest_ensure_response( array(
			'status'		=> 13,
			'crypt_data'	=> array(
				'count'			=> count( $list ),
				'all_pages'		=> $all_pages,
				'items'			=> $list
			),
		) );
	}


	public function callback_subsets_list( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}

		$list		= array();
		$paged		= isset( $data['paged'] ) ? absint( $data['paged'] ) : 1;
		$count		= isset( $data['count'] ) ? absint( $data['count'] ) : 10;
		$start		= intval( ($paged - 1) * $count );
		$results	= mlmFire()->db->query_rows(
			"SELECT user_id, date FROM {TABLE} WHERE parent_id = %d ORDER BY id DESC LIMIT %d, %d",
			array( $user_id, $start, $count ),
			'network'
		);
		$all_pages	= mlmFire()->db->count_query_rows(
			"SELECT COUNT(id) FROM {TABLE} WHERE parent_id = %d",
			array( $user_id ),
			'network'
		);

		if( ! empty( $results ) )
		{
			foreach( $results as $item )
			{
				$list[]		= array(
					'id'			=> $item->id,
					'avatar'		=> get_avatar_url( $item->user_id ),
					'name'			=> mlm_get_user_name( $item->user_id ),
					'refs_count'	=> mlmFire()->referral->get_refs_count( $item->user_id ),
					'subs_count'	=> mlmFire()->network->get_subs_count( $item->user_id ),
					'date'			=> $item->date,
					'local_date'	=> date_i18n( 'Y/m/d H:i:s', strtotime( $item->date ) ),
				);
			}
		}

		return rest_ensure_response( array(
			'status'		=> 13,
			'crypt_data'	=> array(
				'count'			=> count( $list ),
				'all_pages'		=> $all_pages,
				'items'			=> $list
			),
		) );
	}
	
	
	public function callback_order_attach( $data )
	{
		$user_id	= $this->find_user_id_by_token();

		if( ! $user_id )
		{
			return $this->get_error('not_logged_in');
		}
		
		$order_id	= isset( $data['order_id'] ) ? absint( $data['order_id'] ) : 0;
		$customer	= (int)get_post_meta( $order_id, '_customer_user', true );
		$updated	= false;
		
		if( ! $customer )
		{
			$updated	= true;
			
			update_post_meta( $order_id, '_customer_user', $user_id );
		}
		
		return rest_ensure_response( array(
			'status'		=> 13,
			'crypt_data'	=> array(
				'updated'		=> $updated
			),
		) );
	}


	/**
	 * Get WP Error object by error code.
	 */
	private function get_error( $code )
	{
		switch( $code )
		{
			case 'invalid_token':
				$text	= 'توکن نامعتبر است.';
				break;

			case 'already_upgraded':
				$text	= 'حساب شما نیازی به ارتقا ندارد. همه امکانات پنل کاربری برای شما فعال می باشد.';
				break;

			case 'already_verified':
				$text	= 'درخواست ارتقا حساب شما قبلا تایید شده است.';
				break;

			case 'wait_for_result':
				$text	= 'درخواست ارتقا حساب شما در حال بررسی می  باشد.';
				break;

			case 'wrong_blog':
				$text	= 'شما عضو این سایت نیستید.';
				break;

			case 'create_token':
				$text	= 'خطا هنگام ایجاد توکن، لطفا دوباره امتحان کنید.';
				break;

			case 'invalid_login_password':
				$text	= 'نام کاربری یا رمز عبور نادرست است.';
				break;

			case 'invalid_login':
				$text	= 'نام کاربری، موبایل یا ایمیل وارد شده برای هیچ کاربری ثبت نشده است.';
				break;

			case 'invalid_username':
				$text	= 'برای نام کاربری فقط می توانید ترکیبی از اعداد و حروف لاتین استفاده نمایید.';
				break;

			case 'username_exists':
				$text	= 'این نام کاربری قبلا رزرو شده است. لطفا نام کاربری دیگری انتخاب نمایید.';
				break;

			case 'invalid_plan_id':
				$text	= 'شناسه اشتراک معتبر نیست.';
				break;

			case 'invalid_plan_time':
				$text	= 'مدت اعتبار پلن انتخابی قابل قبول نیست. لطفا با پشتیبانی سایت تماس بگیرید.';
				break;

			case 'balance_less_than_amount':
				$text	= 'موجودی کیف پول شما از مبلغ اشتراک کمتر است. لطفا کیف پول خود را شارژ نمایید.';
				break;

			case 'not_logged_in':
				$text	= 'جهت نمایش این قسمت ورود به حساب کاربری الزامی است.';
				break;

			case 'token_expired':
				$text	= 'زمان استفاده از توکن به اتمام رسیده است.';
				break;

			case 'amount_less_than_minimum':
				$text	= 'مبلغ درخواستی از حداقل وجه قابل برداشت کمتر است.';
				break;

			case 'amount_less_than_allowed':
				$text	= 'حداقل مبلغ قابل قبول 5000 تومان می باشد.';
				break;

			case 'signature_error':
				$text	= 'اعتبار توکن تایید نشد.';
				break;

			case 'wrong_token':
				$text	= 'توکن نامعتبر است.';
				break;

			case 'empty_login_pass':
				$text	= 'درج نام کاربری و رمز عبور الزامی است.';
				break;

			case 'empty_mobile':
				$text	= 'درج شماره موبایل معتبر الزامی است.';
				break;

			case 'invalid_mobile':
				$text	= 'شماره موبایل وارد شده معتبر نیست. شماره را به صورت عددی 11 رقمی با شروع از 09 وارد کنید.';
				break;

			case 'mobile_exists':
				$text	= 'این شماره موبایل قبلا ثبت شده است.';
				break;

			case 'mobile_not_exists':
				$text	= 'این شماره موبایل برای هیچ کاربری ثبت نشده است.';
				break;

			case 'wait_for_code':
				$text	= 'کد تایید برای شما ارسال شده است. لطفا چند دقیقه منتظر باشید و در صورت عدم دریافت کد دوباره امتحان کنید.';
				break;

			case 'sms_failed':
				$text	= 'ارسال پیامک با خطا مواجه شد. لطفا دوباره امتحان کنید یا با پشتیبانی سایت در تماس باشید.';
				break;

			case 'required_fields':
				$text	= 'تکمیل فیلدهای ستاره دار الزامی است.';
				break;

			case 'no_verify_code':
				$text	= 'کد تایید برای شما ارسال نشده است. ابتدا دکمه دریافت کد تایید را بزنید.';
				break;

			case 'code_mismatch':
				$text	= 'کد تایید وارد شده نادرست است. کد را با دقت بیشتری وارد کنید.';
				break;

			case 'password_lenght':
				$text	= 'رمز عبور باید حداقل 7 کاراکتر داشته باشد.';
				break;

			case 'invalid_email':
				$text	= 'آدرس ایمیل وارد شده معتبر نیست.';
				break;

			case 'email_exists':
				$text	= 'این آدرس ایمیل برای کاربر دیگری ثبت شده است.';
				break;

			case 'no_items_found':
				$text	= 'هیچ موردی یافت نشد.';
				break;

			case 'insufficient_funds':
				$text	= 'موجودی شما از حداقل وجه قابل برداشت کمتر است.';
				break;

			case 'more_than_balance':
				$text	= 'مبلغ درخواستی از موجودی حساب شما بیشتر است.';
				break;

			case 'pending_withdrawal':
				$text	= 'تا تعیین وضعیت درخواست قبلی امکان ارسال درخواست وجه جدید وجود ندارد.';
				break;

			case 'invalid_contact_id':
				$text	= 'کد دریافت کننده معتبر نیست.';
				break;

			case 'no_ticket_for_yourself':
				$text	= 'شما نمی توانید برای خودتان تیکت ارسال کنید.';
				break;

			case 'subject_required':
				$text = 'لطفا موضوع اصلی را انتخاب کنید.';
				break;

			case 'related_product_required':
				$text = 'انتخاب محصول مرتبط به تیکت الزامی است.';
				break;

			case 'related_department_required':
				$text = 'انتخاب دپارتمان مرتبط به تیکت الزامی است.';
				break;

			case 'invalid_ticket_id':
				$text	= 'کد تیکت معتبر نیست یا حذف شده است.';
				break;

			case 'invalid_gateway':
				$text	= 'درگاه پرداخت انتخاب شده معتبر نیست.';
				break;

			case 'invalid_post_id':
				$text	= 'شناسه نوشته نامعتبر است.';
				break;

			case 'invalid_user_id':
				$text	= 'شناسه کاربر نامعتبر است.';
				break;

			case 'cant_follow_yourself':
				$text	= 'شما نمی توانید خودتان را دنبال کنید.';
				break;

			case 'not_allowed':
				$text	= 'شما مجاز به این کار نیستید.';
				break;

			case 'file_wrong_format':
				$allowed	= array( 'jpg', 'jpeg', 'png', 'gif', 'pdf', 'doc', 'docx', 'zip' );
				$formats	= implode( ' - ', $allowed );
				$text		= 'فرمت های مجاز برای آپلود ' . $formats . ' می باشند.';
				break;

			case 'file_too_large':
				$text	= 'ححداکثر حجم قابل آپلود 5 مگابایت می باشد.';
				break;

			case 'cant_charge_with_wallet':
				$text = 'شما نمی توانید برای شارژ کیف پول از موجودی کیف پول استفاده کنید.';
				break;

			case 'insufficient_fund':
				$text = 'موجودی شما جهت پرداخت کافی نیست. لطفا ابتدا کیف پول خود را شارژ کنید.';
				break;

			case 'no_file_uploaded':
				$text = 'هیچ فایلی برای این محصول انتخاب نشده است.';
				break;

			default:
				$text	= 'خطای غیرمنتظره ای پیش آمده است. لطفا دوباره امتحان کنید.';
				break;
		}

		return new WP_Error( 'mlm_api_auth_'.$code, $text, array( 'status' => 401 ) );
	}
}